!=================================================================
!=================================================================
!=================================================================
!      =====                                           =====
!      =====             MODULE DEM  Class             =====
!      =====                                           =====
!=================================================================
!=================================================================
!=================================================================

!=======================================================================
! This module provides routines for processing DEM terrain elevations
! from the RENCI group.
!
! Revision history:
!    Date		Programmer					Description of change
!    ----		----------					---------------------
!    03/20/07	Craig  Mattocks, UNC-IMS	Wrote original code (Lidar Class)
!    10/04/07   Crystal Fulcher, UNC-IMS    Modified Craig's Lidar code with
!                                           his assistance to use with the
!                                           10m DEM generated at RENCI
!     11/07     Mark S.C. Reed, UNC-RENCI MPI version of code to run in parallel
!=======================================================================
MODULE DEMClass

	!------------------------
	! Import global constants
	!------------------------
	USE ConstantsClass

	!------------------
	! Import File class
	!------------------
	USE FileClass

	!--------------------------------------------
	! Import Triangle class to represent elements
	!--------------------------------------------
	USE TriangleClass

	!--------------------------------------------
	! Force explicit declaration of all variables
	!--------------------------------------------
	IMPLICIT none

	!-----------------------
	! Set access permissions
	!-----------------------
	PUBLIC  ::	userInput, readGrid, writeGrid,									&
					newNeighbors, deleteNeighbors,									&
					findNeighbors, printNeighbors, buildGridConnections,		&
					findPatch, readDEM
	PRIVATE :: loadGrid, dumpGrid, loadDEM

	!-----------------------------------------------------------
	! Temporary arrays to get around Intel compiler inadequacies
	!-----------------------------------------------------------
	INTEGER										:: nnT
	INTEGER										:: neT
	INTEGER,	DIMENSION(:),	ALLOCATABLE	:: nnTemp
	INTEGER,	DIMENSION(:),	ALLOCATABLE	:: neTemp

	!--------------------------------------------------------------
	! Retain values of class instance variables between invocations
	!--------------------------------------------------------------
	SAVE

	!-------------------------
	! Class instance variables
	!-------------------------

	!----------------------
	! ADCIRC grid variables
	!----------------------
	INTEGER,		PARAMETER									:: maxHeader = 256	! Maximum number of characters in file header
	CHARACTER*(maxHeader)									:: gridFileHeader		! Grid file header
	INTEGER														:: np						! Number of grid nodes
	INTEGER														:: ne						! Number of grid elements
	TYPE (Point),		DIMENSION(:),		ALLOCATABLE	:: node					! Grid nodes
	TYPE (Triangle),	DIMENSION(:),		ALLOCATABLE	:: element				! Grid elements
	REAL*8,				DIMENSION(:),		ALLOCATABLE :: zold					! Original depth/height of grid nodes

	!--------------------------------
	! Neighbors of nodes and elements
	!--------------------------------
	INTEGER, PARAMETER							:: maxNodeConnections = 16		! 1 + max number of nodes
																								! connected to any one node
																								! in the finite element grid

	INTEGER											:: minNeighbors					! 1 + minimum number of
																								! neighbors for any node

	INTEGER											:: maxNeighbors					! 1 + maximum number of
																								! neighbors for any node

	INTEGER											:: nne								! Number of neighboring elements

	INTEGER,	DIMENSION(:,:),	ALLOCATABLE	:: elementNodeNumbers			! Node numbers associated
																								! with each element -
																								! the 3 point IDs

	INTEGER,	DIMENSION(:,:),	ALLOCATABLE	:: neighborNodes					! 2D array of neighbors
																								! for each node

	INTEGER,	DIMENSION(:,:),	ALLOCATABLE	:: neighborElements				! 2D array of neighbor
																								! elements for each node

	INTEGER,	DIMENSION(:),		ALLOCATABLE	:: nNeighbors						! Number of neighbors
																								! for each node

	INTEGER,	DIMENSION(:),		ALLOCATABLE	:: neighboringElements			! Neighboring elements
!	INTEGER,	DIMENSION(16)		:: neighboringElements			! Neighboring elements

	REAL(8),	DIMENSION(:),		ALLOCATABLE	:: x, y								! Node coordinates

	REAL(8)											:: sumArea	= zero				! Sum of neighboring elements' areas
	REAL(8)											:: avgZ		= zero				! Area-weighted average of
																								! neighboring elements' z values

        REAL(8), DIMENSION(:), ALLOCATABLE :: avgZarray        ! store average
        ! z values from all processes in an array. Write them out at end.


	!---------------------
	! DEM data variables
	!---------------------
!	CHARACTER*28,		DIMENSION(:),		ALLOCATABLE :: tileName				! List   of LiDAR data tiles
!	INTEGER														:: nTiles				! Number of LiDAR data tiles found
	INTEGER														:: nDEM				! Number of non-missing DEM
																								! data values read in
	REAL*8,				DIMENSION(:),		ALLOCATABLE :: xDEM				! x coordinates of DEM data values
	REAL*8,				DIMENSION(:),		ALLOCATABLE :: yDEM				! y coordinates of DEM data values
	REAL*8,				DIMENSION(:),		ALLOCATABLE :: zDEM				! DEM topo values
	INTEGER														:: nx						! Number of DEM data values
                                                   							! in x direction 
	INTEGER														:: ny						! Number of DEM data values
                                                   							! in y direction  
!	INTEGER														:: nxy					! Maximum number of LiDAR data values
																								! per tile that could be read in
	REAL*8														:: x0						! x origin of DEM
                                                   							! (lower lefthand corner)
	REAL*8														:: y0						! y origin of DEM
                                                   							! (lower lefthand corner)
	REAL*8														:: dxy					! Cell spacing	of DEM
	INTEGER														:: nOff = 0				! Number of datum offset values
	REAL*8														:: zOff = zero			! Datum offset
	REAL*8                                                        :: dem_north
	REAL*8                                                        :: dem_south
	REAL*8                                                        :: dem_west
	REAL*8                                                        :: dem_east

	!---------------
	! File variables
	!---------------
	CHARACTER*80	:: inputGridFile	! Name of input  grid file
	CHARACTER*80	:: outputGridFile	! Name of output grid file
	CHARACTER*80	:: asciiDEM	        ! Ascii DEM file - uses header
	CHARACTER*80  :: dir_accessDEM    ! Direct Access DEM file
	LOGICAL			:: isFileOpen		! File open successful?

	!----------------
	! Class functions
	!----------------
	CONTAINS


		!----------------------------
		! Prompt user to obtain input
		!----------------------------
		SUBROUTINE userInput(inputGridFile, outputGridFile, asciiDEM, dir_accessDEM)
			CHARACTER (*), INTENT(OUT) :: inputGridFile	! Name of input  grid file
			CHARACTER (*), INTENT(OUT) :: outputGridFile	! Name of output grid file
			CHARACTER (*), INTENT(OUT) :: asciiDEM	! Name of ascii DEM file
			CHARACTER (*), INTENT(OUT) :: dir_accessDEM	! Name of direct access DEM file

			WRITE (*, "('Enter the filename for the input  grid: ')")
			READ  (*, "(a)") inputGridFile
			WRITE (*, "(a)") inputGridFile

			WRITE (*, "('Enter the filename for the output grid: ')")
			READ  (*, "(a)") outputGridFile
			WRITE (*, "(a)") outputGridFile

			WRITE (*, "('Enter the filename of the ascii DEM: ')")
			READ  (*, "(a)") asciiDEM
			WRITE (*, "(a)") asciiDEM
			
			WRITE (*, "('Enter the filename of the Direct Access DEM: ')")
			READ  (*, "(a)") dir_accessDEM
			WRITE (*, "(a)") dir_accessDEM
			
		END SUBROUTINE userInput

		!-------------------------
		! Read an ADCIRC grid file
		!-------------------------
		SUBROUTINE readGrid(gridFilename,outgridFilename)
			CHARACTER (*), INTENT(IN) :: gridFilename	! Name of input grid file
			CHARACTER (*), INTENT(IN) :: outgridFilename	! Name of input grid file

			INTEGER :: gridFileUnit							! Logical unit of grid file
			LOGICAL :: isOpen									! File open successful?



			!--------------------------------------------
			! Check out a logical unit for the input file
			!--------------------------------------------
			gridFileUnit = getUnit()

			!---------------------
			! Open input data file
			!---------------------
			isOpen = openReadFile(gridFileUnit, gridFilename)

			!-------------------------
			! Was the OPEN successful? 
			!-------------------------
			IF (isOpen) THEN										! Open successful

				!---------------------------------
				! Load the grid nodes and elements
				!---------------------------------
				CALL loadGrid(gridFileUnit,outgridFilename)

			ELSE														! File open failed

!				WRITE (*,*)	'Subroutine readGrid:  ',								&
!								'Failed to open file ', gridFilename

			END IF	! (isOpen)

			!--------------------------------------------------
			! Close input data file and return its logical unit
			!--------------------------------------------------
			CALL releaseFile(gridFileUnit)	

		END SUBROUTINE readGrid

		!--------------------------------------------
		! Load in the values from an ADCIRC grid file
		!--------------------------------------------
		SUBROUTINE loadGrid(gridFileUnit,gridFilename)
			INTEGER, INTENT(IN)	:: gridFileUnit			! Logical unit of grid file
			CHARACTER (*), INTENT(IN) :: gridFilename	! Name of output grid file
			INTEGER					:: anyErrors				! Status: 0 for success, else failure
			INTEGER :: outgridFileUnit							! Logical unit of output grid file
			REAL*8					:: xx,yy,zz					! Node coordinates and depth/height
			INTEGER					:: num						! Number of nodes per element
			INTEGER					:: n, n1,n2,n3				! Node numbers
			INTEGER					:: i							! Loop index


!			----------------
!			Read file header
!			----------------
			READ (gridFileUnit,'(a)', IOSTAT=anyErrors) gridFileHeader

			IF (anyErrors /= 0) THEN
				WRITE (*,*)	"SUBROUTINE loadGrid: ",								&
	&							"Problem reading grid file header"
				CALL checkFileStatus(anyErrors)
			END IF

!			------------------------------------------------
!			Read in number of ADCIRC grid elements and nodes
!			------------------------------------------------
			READ (gridFileUnit, *, IOSTAT=anyErrors) ne, np

			IF (anyErrors == 0) THEN
				WRITE (*,'("SUBROUTINE loadGrid: ADCIRC grid contains ",		&
	&							i7," nodes and ", i7, " elements.")') np, ne
			ELSE
				WRITE (*,*)	"SUBROUTINE loadGrid: ",								&
	&							"Problem reading # of nodes and elements!"
				CALL checkFileStatus(anyErrors)
			END IF

!			-------------------------------------------
!			Allocate memory for node and element arrays
!			-------------------------------------------
			IF (ALLOCATED(node))		DEALLOCATE (node)
			IF (ALLOCATED(element))	DEALLOCATE (element)
			IF (ALLOCATED(zold))		DEALLOCATE (zold)
			IF (ALLOCATED(avgZarray))		DEALLOCATE (avgZarray)
			ALLOCATE (node(np))
			ALLOCATE (element(ne))
			ALLOCATE (zold(np))
			ALLOCATE (avgZarray(np))
			
			ALLOCATE(elementNodeNumbers(ne,3))
		  	ALLOCATE(neighborNodes(np,maxNodeConnections))
			ALLOCATE(neighborElements(np,maxNodeConnections))
			ALLOCATE(nNeighbors(np))
			ALLOCATE(x(np))
			ALLOCATE(y(np))

! Uncommented these
!			ALLOCATE (angle(maxNodeConnections))
!			ALLOCATE (neiTemp(np))
!			ALLOCATE (nNeighorElements(np))
			
			

!			----------------------
!			Read ADCIRC grid nodes
!			----------------------
			DO i = 1, np
				READ (gridFileUnit, *, IOSTAT=anyErrors) n, xx,yy,zz

				IF (anyErrors == 0) THEN
!					xx = xx + 2500000.d0						! Artificial offset -
!					yy = yy +  500000.d0						! for testing only!!!

         		node(i) = newPoint(n, xx,yy,zero)	! Create node
!					zold(i) = -m2ft * zz						! Save old depth:
																	! -meters --> survey feet
				ELSE
					WRITE (*,*)	"SUBROUTINE loadGrid: ",							&
	&								"Problem reading node #", i
					CALL checkFileStatus(anyErrors)
					EXIT											! Jump out of DO loop
				END IF
			END DO

!			-------------------------
!			Read ADCIRC grid elements
!			-------------------------
			DO i = 1, ne
				READ (gridFileUnit, *, IOSTAT=anyErrors) n,num, n1,n2,n3

				IF (anyErrors == 0) THEN
					element(i) = newTriangle(n, node(n1),node(n2),node(n3))
				ELSE
					WRITE (*,*)	"SUBROUTINE loadGrid: ",							&
	&								"Problem reading element #", i
					CALL checkFileStatus(anyErrors)
					EXIT											! Jump out of DO loop
				END IF
			END DO

!  This part is patched in to get a working version of the code
       		!---------------------------------------------
			! Check out a logical unit for the output file
			!---------------------------------------------
			outgridFileUnit = getUnit()

			!----------------------
			! Open output data file
			!----------------------
			
			OPEN (UNIT=outgridFileUnit, FILE=gridFilename, STATUS='UNKNOWN', ACTION='WRITE')
			
			
!			isOpen = openWriteFile(gridFileUnit,gridFilename)
!			isOpen = openWriteFile(gridFileUnit,gridFilename, maxHeader)

				open(25,file='bathy_topo.out')


!  End of patch


		END SUBROUTINE loadGrid

		!------------------------------
		! Write out an ADCIRC grid file
		!------------------------------
		SUBROUTINE writeGrid(gridFilename)
			CHARACTER (*), INTENT(IN) :: gridFilename	! Name of output grid file

			INTEGER :: gridFileUnit							! Logical unit of output grid file
			LOGICAL :: isOpen									! File open successful?

			!---------------------------------------------
			! Check out a logical unit for the output file
			!---------------------------------------------
			gridFileUnit = getUnit()

			!----------------------
			! Open output data file
			!----------------------
			
			OPEN (UNIT=gridFileUnit, FILE=gridFilename, STATUS='UNKNOWN', ACTION='WRITE')
			
			
!			isOpen = openWriteFile(gridFileUnit,gridFilename)
!			isOpen = openWriteFile(gridFileUnit,gridFilename, maxHeader)

			!-------------------------
			! Was the OPEN successful? 
			!-------------------------
			IF (isOpen) THEN										! Open successful
				!-----------------------------------------------
				! Write the grid nodes and elements to grid file
				!-----------------------------------------------
				CALL dumpGrid(gridFileUnit)

			ELSE														! File open failed

				WRITE (*,*)	'Subroutine writeGrid:  ',								&
								'Failed to open file ', gridFilename

			END IF	! (isOpen)

			!---------------------------------------------------
			! Close output data file and return its logical unit
			!---------------------------------------------------
			CALL releaseFile(gridFileUnit)

		END SUBROUTINE writeGrid

		!--------------------------------------------
		! Write out the values in an ADCIRC grid file
		!--------------------------------------------
		SUBROUTINE dumpGrid(gridFileUnit)
			INTEGER, INTENT(IN)	:: gridFileUnit	! Logical unit of grid file
			INTEGER					:: i					! Loop index

!			-----------------
!			Write file header
!			-----------------
			WRITE (gridFileUnit, '(a)') ADJUSTL(TRIM(gridFileHeader))

!			----------------------------------------------
!			Write number of ADCIRC grid elements and nodes
!			----------------------------------------------
			WRITE (gridFileUnit, '(i6, i8)') ne, np

!			-----------------------
!			Write ADCIRC grid nodes
!			-----------------------
			DO i = 1, np
!         	CALL convertZ(node(i), -ft2m)			! survey feet --> -meters
				CALL writePoint(node(i), gridFileUnit)
			END DO

!			--------------------------
!			Write ADCIRC grid elements
!			--------------------------
			DO i = 1, ne
				CALL writeTriangle (element(i), gridFileUnit)
			END DO
		END SUBROUTINE dumpGrid

		!-----------------------
		! Read a DEM bathy/topo file
		!-----------------------
		SUBROUTINE readDEM(asciiDEM,DEMFileUnit) 
			CHARACTER (*), INTENT(IN)	:: asciiDEM    	! Name of input LiDAR topo file
			INTEGER, INTENT(OUT):: DEMFileUnit					! Logical unit of DEM bath/topo file
			INTEGER							:: asciiDEMFileUnit	! Logical unit of LiDAR topo file
			LOGICAL							:: isOpen			! File open successful?


			!--------------------------------------------
			! Check out a logical unit for the input file
			!--------------------------------------------
			asciiDEMFileUnit = getUnit()

			!---------------------
			! Open input data file
			!---------------------
			isOpen = openReadFile(asciiDEMFileUnit, asciiDEM)
			!-------------------------
			! Was the OPEN successful? 
			!-------------------------
			IF (isOpen) THEN										! Open successful

				!---------------------------------
				! Load the header of the ascii grid file
				!---------------------------------
				CALL loadDEM(asciiDEMFileUnit,DEMFileUnit)

			ELSE														! File open failed

!				WRITE (*,*)	'Subroutine readDEM:  ',								&
!								'Failed to open file ', gridFilename
			   stop

			END IF	! (isOpen)
			

			!--------------------------------------------------
			! Close input data file and return its logical unit
			!--------------------------------------------------
			CALL releaseFile(asciiDEMFileUnit)	

		END SUBROUTINE readDEM

		!------------------------------------------
		! Load in the header from the Ascii DEM file
		! and open the Direct Access DEM file
		!------------------------------------------
		SUBROUTINE loadDEM(asciiDEMFileUnit,DEMFileUnit)
			INTEGER, INTENT(IN)	:: asciiDEMFileUnit					! Logical unit of DEM bath/topo file
			INTEGER, INTENT(OUT):: DEMFileUnit					! Logical unit of DEM bath/topo file
			INTEGER					:: anyErrors						! Status: 0 for success, else failure

			!-----------------
			! Read file header
			!-----------------

! ?? replace these reads!!			
         READ (asciiDEMFileUnit, "(6x,i5)", IOSTAT=anyErrors) nx
         READ (asciiDEMFileUnit, "(6x,i5)", IOSTAT=anyErrors) ny
         READ (asciiDEMFileUnit, "(10x,f10.6)", IOSTAT=anyErrors) x0    
         READ (asciiDEMFileUnit, "(10x,f9.6)", IOSTAT=anyErrors) y0    
         READ (asciiDEMFileUnit, "(9x,f8.6)", IOSTAT=anyErrors) dxy    
         READ (asciiDEMFileUnit, "(13x,i5)", IOSTAT=anyErrors) missing
  
!   Hardwiring this because of truncation problem with current version of the DEM	   
	   dxy=0.0000925925925925925
         
         dem_north=y0+(ny-1)*dxy
         dem_south=y0
         dem_west=x0
         dem_east=x0+(nx-1)*dxy
         
	   DEMFileUnit = getUnit()

         
!         open(DEMFileUnit, file=dir_accessDEM, status='old', access='direct', recl=nx*4)
         open(DEMFileUnit, file=dir_accessDEM, status='old', access='direct', recl=nx*4, form='unformatted')
         
         
         END SUBROUTINE loadDEM
         
         
		!----------------
		! Grid destructor
		!----------------
		SUBROUTINE deleteGrid()
			INTEGER :: i													! Loop index

			DO i = 1, np
				CALL deletePoint(node(i))
			END DO

			DO i = 1, ne
				CALL deleteTriangle(element(i))
			END DO

			IF (ALLOCATED(zold))		DEALLOCATE (zold)
			IF (ALLOCATED(element))	DEALLOCATE (element)
			IF (ALLOCATED(node))		DEALLOCATE (node)
		END SUBROUTINE deleteGrid

		!=================================================================
		! Subroutine to generate neighbor tables from a connectivity table.
		!
		! 	Notes:
		!		1.	A node neighbor table is generated with the node itself
		!			is listed as neighbor #1. All other neighbors are sorted
		!			and placed in clockwise order from East.
		!		2.	A neighbor element table is generated with:
		!			entry 1 = element # defined by neighbors 1,2,3
		!			entry 2 = element # defined by neighbors 1,3,4
		!			entry 3 = element # defined by neighbors 1,4,5
		!			. . .
		!			entry last = element # defined by neighbors 1,nNeighbors,2
		!			A zero area means that the defined triangle lies outside
		!			the domain.
		!
		! Revision history:
		!    Date		Programmer					Description of change
		!    ----		----------					---------------------
		!    06/29/99	Rick Luettich, UNC-IMS	Used in 3D ADCIRC code
		!    05/23/02	Rick Luettich, UNC-IMS	Adapted to provide neighbor
		!													element table 
		!    03/30/07	Craig Mattocks, UNC-IMS	Revamped for bathy/topo
		!													processing
		!=================================================================
		SUBROUTINE findNeighbors()

			IMPLICIT none

			INTEGER :: n,nn, i,j,k, jj,jlow					! Loop indices

			INTEGER :: nn1,nn2,nn3
			INTEGER :: ne1,ne2,ne3
			REAL(8) :: deltaX,deltaY, dist
			REAL(8) :: angleLow, angleMore
			REAL(8) :: angle(maxNodeConnections)
			INTEGER :: neiTemp(np)
			INTEGER :: nNeighorElements(np)

			!--------------------------------------
			! Allocate memory for local work arrays
			!--------------------------------------
!			IF (.NOT. ALLOCATED(angle))				ALLOCATE (angle(maxNodeConnections))
!			IF (.NOT. ALLOCATED(neiTemp))				ALLOCATE (neiTemp(np))
!			IF (.NOT. ALLOCATED(nNeighorElements))	ALLOCATE (nNeighorElements(np))

			nNeighbors			= 0
			nNeighorElements	= 0
			neighborNodes		= 0
			neighborElements	= 0

			DO 10 n = 1, ne
				nn1 = elementNodeNumbers(n,1)
				nn2 = elementNodeNumbers(n,2)
				nn3 = elementNodeNumbers(n,3)

				nNeighorElements(nn1) = nNeighorElements(nn1) + 1
				nNeighorElements(nn2) = nNeighorElements(nn2) + 1
				nNeighorElements(nn3) = nNeighorElements(nn3) + 1

				neighborElements(nn1, nNeighorElements(nn1)) = n
				neighborElements(nn2, nNeighorElements(nn2)) = n
				neighborElements(nn3, nNeighorElements(nn3)) = n

				DO j = 1, nNeighbors(nn1)
					IF (nn2 == neighborNodes(nn1,j)) GO TO 25
				END DO

				nNeighbors(nn1) = nNeighbors(nn1) + 1
				nNeighbors(nn2) = nNeighbors(nn2) + 1
				IF ((nNeighbors(nn1) > maxNodeConnections-1) .OR.				&
					 (nNeighbors(nn2) > maxNodeConnections-1))					&
					GO TO 999
				neighborNodes(nn1, nNeighbors(nn1)) = nn2
				neighborNodes(nn2, nNeighbors(nn2)) = nn1

 25			CONTINUE

				DO j = 1, nNeighbors(nn1)
					IF (nn3 == neighborNodes(nn1,j)) GO TO 35
				END DO

				nNeighbors(nn1) = nNeighbors(nn1) + 1
				nNeighbors(nn3) = nNeighbors(nn3) + 1
				IF ((nNeighbors(nn1) > maxNodeConnections-1) .OR.				&
					 (nNeighbors(nn3) > maxNodeConnections-1))					&
					GO TO 999
				neighborNodes(nn1, nNeighbors(nn1)) = nn3
				neighborNodes(nn3, nNeighbors(nn3)) = nn1

35				CONTINUE

				DO j = 1, nNeighbors(nn2)
					IF (nn3 == neighborNodes(nn2,j)) GO TO 10
				END DO

				nNeighbors(nn2) = nNeighbors(nn2) + 1
				nNeighbors(nn3) = nNeighbors(nn3) + 1
				IF ((nNeighbors(nn2) > maxNodeConnections-1) .OR.				&
					 (nNeighbors(nn3) > maxNodeConnections-1))					&
					GO TO 999
				neighborNodes(nn2, nNeighbors(nn2)) = nn3
				neighborNodes(nn3, nNeighbors(nn3)) = nn2

10			CONTINUE

			!--------------------------------------------------------
			! Insert node itself in place #1 and sort other neighbors
			! by increasing clockwise angle from East
			!--------------------------------------------------------
			DO i = 1,np
				DO j = 1, nNeighbors(i)
					neiTemp(j) = neighborNodes(i,j)

					deltaX = x(neiTemp(j)) - x(i)
					deltaY = y(neiTemp(j)) - y(i)

					dist = SQRT(deltaX*deltaX + deltaY*deltaY)
					IF (dist == zero) GO TO 998

					IF (deltaY /= zero) THEN
						angle(j) = rad2deg * ACOS(deltaX/dist)
						IF (deltaY > zero) angle(j) = threeSixty - angle(j)
					END IF

					IF (deltaY == zero) THEN
						IF (deltaX > zero) angle(j) = zero
						IF (deltaX < zero) angle(j) = oneEighty
					END IF
				END DO	! j = 1, nNeighbors(i)

				angleMore = -one

				DO jj = 1, nNeighbors(i)
					angleLow = fourHundred
					DO j = 1, nNeighbors(i)
						IF ((angle(j) < angleLow) .AND.								&
							 (angle(j) > angleMore)) THEN
							angleLow = angle(j)
							jlow = j
						END IF
					END DO	! j = 1, nNeighbors(i)

					neighborNodes(i,jj+1) = neiTemp(jlow)
					angleMore = angleLow
				END DO	! jj = 1, nNeighbors(i)

				neighborNodes(i,1) = i
				nNeighbors(i) = nNeighbors(i) + 1
			END DO	! i = 1,np

			!-----------------------------------------------------
			! Match each set of 3 nodes with corresponding element
			! and reorder elements accordingly
			!-----------------------------------------------------
			DO i = 1, np
				DO k = 1, nNeighorElements(i)
					neiTemp(k) = neighborElements(i,k)
					neighborElements(i,k) = 0
				END DO

				DO j = 2, nNeighbors(i)
					nn1 = neighborNodes(i,1)
					nn3 = neighborNodes(i,j)
					IF (j /= nNeighbors(i)) nn2 = neighborNodes(i,j+1)
					IF (j == nNeighbors(i)) nn2 = neighborNodes(i,  2)

					DO k = 1, nNeighorElements(i)
						IF (neiTemp(k) /= 0) THEN
							IF (elementNodeNumbers(neiTemp(k),1) == nn1) THEN
								ne1 = elementNodeNumbers(neiTemp(k), 1)
								ne2 = elementNodeNumbers(neiTemp(k), 2)
								ne3 = elementNodeNumbers(neiTemp(k), 3)
							END IF

							IF (elementNodeNumbers(neiTemp(k),2) == nn1) THEN
								ne1 = elementNodeNumbers(neiTemp(k), 2)
								ne2 = elementNodeNumbers(neiTemp(k), 3)
								ne3 = elementNodeNumbers(neiTemp(k), 1)
							END IF

							IF (elementNodeNumbers(neiTemp(k),3) == nn1) THEN
								ne1 = elementNodeNumbers(neiTemp(k), 3)
								ne2 = elementNodeNumbers(neiTemp(k), 1)
								ne3 = elementNodeNumbers(neiTemp(k), 2)
							END IF

							IF ((ne2 == nn2) .AND. (ne3 == nn3)) THEN
								neighborElements(i,j-1) = neiTemp(k)
								neiTemp(k) = 0
							END IF
						END IF
					END DO	! k = 1, nNeighorElements(i)
				END DO	! j = 2, nNeighbors(i)
			END DO	! i = 1, np

			!------------------------------------------------------
			! Determine the maximum and minimum number of neighbors
			!------------------------------------------------------
			minNeighbors = MINVAL(nNeighbors)
			maxNeighbors = MAXVAL(nNeighbors)

			!----------------------------
			! Write out diagnostic output
			!----------------------------
!			WRITE (*,'("Minimum number of neighbors is ", i3)') minNeighbors
!			WRITE (*,'("Maximum number of neighbors is ", i3)') maxNeighbors

!			WRITE (*,'("-- INSIDE findNeighbors ---")')
!			DO n=1,np
!				WRITE (*,'("Neighbor nodes and elements for node #", i6, ":")') n
!				WRITE (*,'(10(1x,i7))') (neighborNodes   (n,j), j=1,nNeighbors(n)  )
!				WRITE (*,'(10(1x,i7))') (neighborElements(n,j), j=1,nNeighbors(n)-1)
!			END DO

			!------------------------------------------
			! Deallocate local work arrays.
			! Free up memory off heap in reverse order.
			!------------------------------------------
!			IF (ALLOCATED(nNeighorElements))	DEALLOCATE (nNeighorElements)
!			IF (ALLOCATED(neiTemp))				DEALLOCATE (neiTemp)    
!			IF (ALLOCATED(angle))				DEALLOCATE (angle)     

			RETURN	! Done

			!---------------------------------------------------------------
			! Terminate program if maximum number of neighbors set too small
			!---------------------------------------------------------------
999		WRITE (*,																			&
	&		"(/,' !!! WARNING - FATAL ERROR in subroutine findNeighbors !!!',	&
	&		/,' The dimensioning parameter maxNodeConnections is too small',	&
	&		/,' There is a problem with the dynamic memory allocation',			&
	&		/,' !!!!!! Execution will now be terminated !!!!!!')")
			STOP

998		WRITE (*,																			&
	&		"(/,' !!! WARNING - FATAL ERROR in subroutine findNeighbors !!!',	&
	&		/,' Nodes ',I7,' and ',I7,' have the same coordinates',				&
	&		/,' !!!!!! Execution will now be terminated !!!!!!')")				&
	&		i, neiTemp(j)
			STOP

		END SUBROUTINE findNeighbors

		!----------------------------------------------------
		! Allocate memory for neighbors of nodes and elements
		!----------------------------------------------------
		SUBROUTINE newNeighbors()

				ALLOCATE(elementNodeNumbers(ne,3))
		    

			IF (.NOT. ALLOCATED(neighborNodes))										&
				ALLOCATE(neighborNodes(np,maxNodeConnections))

			IF (.NOT. ALLOCATED(neighborElements))									&
				ALLOCATE(neighborElements(np,maxNodeConnections))

			IF (.NOT. ALLOCATED(nNeighbors))											&
				ALLOCATE(nNeighbors(np))

			IF (.NOT. ALLOCATED(x))														&
				ALLOCATE(x(np))

			IF (.NOT. ALLOCATED(y))														&
				ALLOCATE(y(np))
		END SUBROUTINE newNeighbors

		!------------------------------------------------------------
		! Free memory associated with neighbors of nodes and elements
		!------------------------------------------------------------
		SUBROUTINE deleteNeighbors()
			IF (ALLOCATED(y))							DEALLOCATE(y)
			IF (ALLOCATED(x))							DEALLOCATE(x)
			IF (ALLOCATED(nNeighbors))				DEALLOCATE(nNeighbors)
			IF (ALLOCATED(neighborElements))		DEALLOCATE(neighborElements)
			IF (ALLOCATED(neighborNodes))			DEALLOCATE(neighborNodes)
			IF (ALLOCATED(elementNodeNumbers))	DEALLOCATE(elementNodeNumbers)
		END SUBROUTINE deleteNeighbors

		!------------------------------------------
		! Find neighbors of grid nodes and elements
		!------------------------------------------
		SUBROUTINE buildGridConnections()
			INTEGER :: n		
			!--------------------------------------
			! Allocate memory for finding neighbors
			!--------------------------------------
!			CALL newNeighbors()

			DO n = 1, np
				x(n) = getX(node(n))
				y(n) = getY(node(n))
			END DO

			DO n = 1, ne
				elementNodeNumbers(n,1) = getID( getP1(element(n)) )
				elementNodeNumbers(n,2) = getID( getP2(element(n)) )
				elementNodeNumbers(n,3) = getID( getP3(element(n)) )
			END DO

			!------------------------------------------
			! Find neighbors of grid nodes and elements
			!------------------------------------------
			CALL findNeighbors()

			!----------------------------------------------
			! Set nodes' neighbors and neighboring elements
			!----------------------------------------------
			DO n = 1, np
				nnT = nNeighbors(n)
				neT = nNeighbors(n) - 1
!			   IF (.NOT. ALLOCATED(nnTemp))	ALLOCATE(nnTemp(nnT))
!			   IF (.NOT. ALLOCATED(neTemp))	ALLOCATE(neTemp(neT))

!				ALLOCATE(nnTemp(nnT))
!				ALLOCATE(neTemp(neT))
!				nnTemp = neighborNodes		(n, 1:nnT)
!				neTemp = neighborElements	(n, 1:neT)

!				CALL setNeighbors(node(n), nnT,neT, nnTemp,neTemp)
				CALL setNeighbors(node(n), nnT,neT, neighborNodes(n,1:nnT),neighborElements(n,1:neT))
				
 ! 		        IF (ALLOCATED(nnTemp))				DEALLOCATE(nnTemp)
  !		        IF (ALLOCATED(neTemp))				DEALLOCATE(neTemp)

!				DEALLOCATE(nnTemp)
!				DEALLOCATE(neTemp)

!				CALL printNeighbors(node(n))
			END DO

			!---------------------------------------
			! Free memory used for finding neighbors
			!---------------------------------------
!			CALL deleteNeighbors()
		END SUBROUTINE buildGridConnections

		!---------------------------------------------------
		! Compute the average (or envelope) terrain height
		! from DEM data for all ADCIRC grid nodes.
		!---------------------------------------------------
		SUBROUTINE computeTerrain(DEMFileUnit)

			!---------------------------------
			! Local LiDAR processing variables
			!---------------------------------
			TYPE (Point)	:: p					! Temporary node
			REAL*8			:: eArea				! Element area
			REAL*8			:: eZ					! Element terrain height
			LOGICAL			:: isFileOpen		! File open successful?
			INTEGER			:: i,j,k, n			! Loop indices
	      CHARACTER*80	:: outputGridFile	! Name of output grid file
			REAL*8          :: north, south, east, west
	      REAL*8,				DIMENSION(:),		ALLOCATABLE :: xDEM				! x coordinates of DEM data values
	      REAL*8,				DIMENSION(:),		ALLOCATABLE :: yDEM				! y coordinates of DEM data values
	      REAL*8,				DIMENSION(:),		ALLOCATABLE :: zDEM				! DEM topo values
	      REAL*8,				DIMENSION(:),		ALLOCATABLE :: ztemp				! DEM topo values
			INTEGER         :: n_row_n,n_row_s,n_col_e,n_col_w
			INTEGER         :: DEMFileUnit
!			INTEGER         :: neighboringElements(16)

                        INTEGER :: mystart, mystop, mycount
                        INTEGER :: nprocs, myrank, ierr, master
                        INTEGER, DIMENSION(:),	ALLOCATABLE :: blocklengths,&
                             & displ, owners

			REAL*8 :: starttime, stoptime, walltime

                        ! MPI include file
                        include "mpif.h"


            ALLOCATE(zDEM(nx*10000),xDEM(nx*10000),yDEM(nx*10000),ztemp(nx*10000))
            ALLOCATE(neighboringElements(16))


            ! find out how many processes are running and who we are
            call MPI_Comm_size(MPI_COMM_WORLD, nprocs, ierr)
            call MPI_Comm_rank(MPI_COMM_WORLD, myrank, ierr)

            ! allocate memory for arrays need for final gather
            IF (ALLOCATED(blocklengths)) DEALLOCATE (blocklengths)
            ALLOCATE (blocklengths(nprocs))
            IF (ALLOCATED(displ)) DEALLOCATE (displ)
            ALLOCATE (displ(nprocs))
            IF (ALLOCATED(owners)) DEALLOCATE (owners)
            ALLOCATE (owners(ne))





! call routine to find out which elements are handled by this process
	    call getDomain(ne, mystart, mystop, blocklengths, displ, owners)

			!-------------------------------------------------
			! Compute the average (or envelope) terrain height
			! for all grid elements.
			!-------------------------------------------------
	                starttime = MPI_Wtime()
			DO n = mystart,mystop	! Loop over grid elements
!			DO n = 1, ne	! Loop over grid elements
!			DO n = 1, 12
				!---------------------------------------
				! Find tiles that cover the grid element
				!---------------------------------------

				CALL findPatch (n,n_row_n,n_row_s,n_col_w,n_col_e)
				!---------------------------------------
				! Read the appropriate rows of data and
				! calculate the x and y values for each
				! point in the patch 
				!---------------------------------------

                nDEM=0
                
                do k=n_row_n,n_row_s
                    read(DEMFileUnit, rec=k) (ztemp(j), j=1,nx)
                     do i=n_col_w,n_col_e
                        if (ztemp(i)==missingDouble) goto 5
                        nDEM=nDEM+1
                        xDEM(nDEM)=dem_west+(i-1)*dxy
                        yDEM(nDEM)=dem_north-(k-1)*dxy
                        zDEM(nDEM)=ztemp(i)
5                       continue                        
                        enddo
                     enddo


			!-------------------------------
			! Free old memory and reallocate
			!-------------------------------
!			IF (ALLOCATED(zDEM)) DEALLOCATE(zDEM)
!			IF (ALLOCATED(yDEM)) DEALLOCATE(yDEM)
!			IF (ALLOCATED(xDEM)) DEALLOCATE(xDEM)
!           ALLOCATE (xDEM(nDEM), yDEM(nDEM), zDEM(nDEM))


					!-------------------------------------------
					! If data is available, "bin" the bathy/topo
					! values that fall inside the element.
					!-------------------------------------------
					
					
					DO i = 1, nDEM
						p = newPoint(i, xDEM(i), yDEM(i), zDEM(i))
						if (zDEM(i).eq.-9999.d0) goto 10

						IF ( isPointNearTriangle (element(n), p) ) THEN
!									WRITE(*,'("LiDAR point #", i6," is NEAR element ", i6)') i,n

									IF ( isPointInTriangle(element(n), p) ) THEN
										CALL addZ(element(n), zDEM(i))
									END IF	! p inside	element

								END IF		! p near		element
10                              continue								
								CALL deletePoint(p)
							END DO	! i = 1, nLidar



				!--------------------------------------
				! Calculate average or envelope z-value
				! over face of grid element.
				!--------------------------------------
				CALL calcZ(element(n))

				IF (MOD(n, 1000) == 0 .AND. getZee(element(n)) /= zero) THEN
					WRITE (*, '(/,i8," LiDAR observations fell in element #", i7)')	&
							getNZ(element(n)), n

					WRITE (*, '("Average z-value for element #", i6," is ",	&
	&								f10.4, " meters")')									&
								n, getZee(element(n)) * ft2m

					WRITE (*, '("and the standard deviation          is ",	&
	&						f10.4, " meters")')											&
							getSdev(element(n)) * ft2m
				END IF

			END DO	! element loop: n = 1, ne
			stoptime = MPI_Wtime()
			walltime = stoptime-starttime
			write (*,1111) "rank= ",myrank," elementLoopTime = ",walltime, " seconds"


			starttime = MPI_Wtime()
                        call exchangeTriangles (ne, element, mystart, mystop, blocklengths, displ)
			stoptime = MPI_Wtime()
			walltime = stoptime-starttime
			write (*,1111) "rank= ",myrank," exchangeTrianglesTime = ",walltime, " seconds"


!!$                        call MPI_Barrier(MPI_COMM_WORLD, ierr)
!!$			! this will surely run like a dog but it should
!!$			! be correct at least
!!$			! loop over every element and broadcast all z values
!!$			DO n = 1, ne	! Loop over grid elements
!!$			   eZ = getZee(element(n))
!!$                           print *, "bcast eZ=",eZ," from ",owners(n)," element ", n
!!$                           call MPI_Bcast (eZ, 1, MPI_DOUBLE_PRECISION, owners(n), MPI_COMM_WORLD, ierr)
!!$                           if (myrank.ne.owners(n)) then
!!$                              CALL addZ(element(n), eZ)
!!$                           endif
!!$                           eZ = getZee(element(n))
!!$                           write (60+myrank,*) " element ", n," calcz= ",eZ
!!$			END DO	! element loop: n = 1, ne
!!$			




			!------------------------------------------
			! Find neighbors of grid nodes and elements
			!------------------------------------------
			CALL buildGridConnections()


			!--------------------------------------------------------------
			! Compute terrain height at each grid node as the area-weighted
			! average of its neighboring elements' z values
			!--------------------------------------------------------------
			nOff = 0
			zOff = zero

! call routine to find out which elements are handled by this process
	    call getDomain(np, mystart, mystop, blocklengths, displ, owners)

			starttime = MPI_Wtime()

!			DO n = 1, np	! Loop over grid nodes
			DO n = mystart,mystop	! Loop over grid nodes
			   do i=1,16
			      neighboringElements(i)=0
			      enddo
!			DO n = 1, 7
				nne = getNumberOfNeighboringShapes(node(n))

!				IF (ALLOCATED(neighboringElements)) DEALLOCATE(neighboringElements)
!				ALLOCATE(neighboringElements(nne))

				CALL getNeighboringShapes(node(n), nne, neighboringElements)
!				write(*,*) 'getNeighboringShapes', n, nne, (neighboringElements(j),j=1,nne)
				

				sumArea	= zero
				avgZ		= zero
				

				DO i = 1, nne
					IF (neighboringElements(i) /= 0) THEN
						IF (getNZ(element(neighboringElements(i))) > 0) THEN
!						    write(*,*) n,i
							eZ			= getZee (element(neighboringElements(i)))
							eArea		= getArea(element(neighboringElements(i)))
							avgZ		= avgZ		+ eArea * eZ
							sumArea	= sumArea	+ eArea
						END IF
					END IF
				END DO

				IF (sumArea > zero) THEN
					avgZ = avgZ / sumArea
				ELSE
					avgZ = zero
				END IF

				!---------------------------------------------------
				! If no LiDAR data at grid node, use original value.
				!---------------------------------------------------
!				IF (avgZ == zero) avgZ = zold(n)

				CALL setZ(node(n), avgZ)	! Set node's z value
! can't write this out here. Save in an array and write out from the master process
!				write(25,*) n, avgZ
                                avgZarray(n) = avgZ


				!---------------------------
				! Accumulate z datum offsets
				!---------------------------
				IF (avgZ /= zero) THEN
!				IF ((avgZ /= zero) .AND. (zold(n) > zGround)) THEN
					nOff = nOff + 1
					zOff = zOff + (avgZ - zold(n))
				END IF

				IF (MOD(n, 1000) == 0 .AND. avgZ /= zero) THEN
					WRITE(*,'("Area-weighted terrain height at node #",		&
	&							i7, " is ", f10.4, " meters")') n, ft2m * avgZ
				END IF

!				IF (ALLOCATED(neighboringElements)) DEALLOCATE(neighboringElements)
			END DO	! node loop: n = 1, np
			stoptime = MPI_Wtime()
			walltime = stoptime-starttime
			write (*,1111) "rank= ",myrank," nodeLoopTime = ",walltime, " seconds"


! First gather the avgZ data from the other nodes
			master = 0
			mycount = mystop-mystart+1
			starttime = MPI_Wtime()
			call mpi_gatherv (avgZarray(mystart),mycount, MPI_DOUBLE_PRECISION, &
                             avgZarray, blocklengths, displ, MPI_DOUBLE_PRECISION, &
                             master, MPI_COMM_WORLD, ierr)
			stoptime = MPI_Wtime()
			walltime = stoptime-starttime
			write (*,1111) "rank= ",myrank," gathervTime = ",walltime, " seconds"


! Loop over grid nodes and write out the output file from the master
                        if (myrank.eq.master) then
                           DO n = 1, np	! Loop over grid nodes
                              write(25,*) n, avgZarray(n)
                           END DO	! node loop: n = 1, np
                        endif



			!-----------------------------------------
			! Compute average z datum offset over land
			!-----------------------------------------
			IF (nOff > 0) THEN
				zOff = zOff / nOff
				WRITE(*,'("Average z datum offset over ", i6,					&
	&						" land points is ", f10.4, " feet or ", f10.4,		&
	&						" meters")') nOff, zOff, (ft2m * zOff)
			END IF
			
		!CALL releaseFile(DEMFileUnit)	
1111	format (a6,i4,a25,f15.2,a8)
		
		END SUBROUTINE computeTerrain
		


! **************************************************************************
! **************************************************************************
                SUBROUTINE getDomain (ne, mystart, mystop, blocklengths,&
                     & displ, owners)

! this computes the domain for each process. It splits the number of elements
! as evenly as it can over all the ranks. Each rank is assigned a contiguous
! block of elements. Some block may have one more element (i.e. longBlocks) 
! than others (shortBlocks) if the number of elements is not evenly divisible
! by the number of processes.
! There is probably a cleaner way to do this but this works so let's declare
! victory and move on. :)

! Note this also returns two arrays used to gather the data
! an array of blocklengths and an array of strides

                  implicit none

                  ! MPI include file
                  include "mpif.h"


                  integer ne        ! the number of elements to decompose 
                  integer mystart   ! the first element for this rank
                  integer mystop    ! the last element for this rank

                  integer  numLongBlocks, numShortBlocks, longBlockSize, shortBlockSize
                  integer myNumLongBlks, myNumShortBlks, myBlockSize
                  integer nprocs, myrank, ierr, i, j, idx
                  INTEGER, DIMENSION(:) :: blocklengths, displ, owners


                  ! first find out our rank and number of processes
                  call MPI_Comm_size(MPI_COMM_WORLD, nprocs, ierr)
                  call MPI_Comm_rank(MPI_COMM_WORLD, myrank, ierr)

                  ! the size of the longer blocks
                  longBlockSize = ceiling(float(ne)/float(nprocs))
                  shortBlockSize = floor(float(ne)/float(nprocs))

                  ! compute the number of blocks of each type
                  numLongBlocks = mod (ne,nprocs)
                  numShortBlocks = nprocs - numLongBlocks


                  
                  ! compute the number of blocks for this rank foreach block type
                  myNumLongBlks = min(numLongBlocks, myrank+1)
                  myNumShortBlks = myrank + 1 - myNumLongBlks

                  ! compute the last element
                  mystop = myNumLongBlks*longBlockSize + myNumShortBlks*shortBlockSize 
                  if (myrank .lt. numLongBlocks) then
                     myBlockSize = longBlockSize 
                  else
                     myBlockSize = shortBlockSize
                  endif
                  ! compute the first element
                  mystart = mystop - myBlockSize + 1




                  ! Now fill the 2 arrays used for gathering data 
                  ! initialize all to longBlockSize then reset the shorter ones
                  blocklengths = longBlockSize
                  do i = numLongBlocks+1, nprocs
                     blocklengths(i) = shortBlockSize
                  enddo

                  ! Now do displacements
                  displ(1) = 0
                  do i = 2, nprocs
                     displ(i) = displ(i-1) + blocklengths(i-1)
                  enddo


                  ! compute who owns each piece
                  ! each rank needs this info
                  idx = 0
                  do i=1,nprocs
                     do j=1,blocklengths(i)
                        ! keep a running count of where we are
                        idx = idx +1
                        ! set the owner equal to the current rank
                        owners(idx) = i-1
                     enddo
                  enddo
                     

		END SUBROUTINE getDomain
! **************************************************************************
! **************************************************************************
SUBROUTINE exchangeTriangles (ne, element, mystart, mystop, blocklengths, displ)

! this routine exchanges values for z, nz, and sdev that were computed for the
! triangle on one processor but need to be used by others.

! 1) gather values to be exchanged into the tri_z, tri_nz, and tri_sdev arrays
! these are all the triangle values that have been modified by the call to addZ
! 2) do an all gather s.t. each processor has all values
! 3) scatter the results back to the triangles using the (new) setter method, setZ


  ! MPI include file
  include "mpif.h"


  integer ne        ! the number of elements 
  integer mystart   ! the first element for this rank
  integer mystop    ! the last element for this rank

  integer nprocs, myrank, ierr, i
  INTEGER, DIMENSION(:) :: blocklengths, displ
  REAL(8), DIMENSION(:), ALLOCATABLE :: tri_z, tri_sdev
  REAL(8), DIMENSION(:), ALLOCATABLE :: alltri_z, alltri_sdev
  INTEGER, DIMENSION(:), ALLOCATABLE :: tri_nz, alltri_nz
  TYPE (Triangle),	DIMENSION(:),		ALLOCATABLE	:: element				! Grid elements





  ! first find out our rank and number of processes
  call MPI_Comm_size(MPI_COMM_WORLD, nprocs, ierr)
  call MPI_Comm_rank(MPI_COMM_WORLD, myrank, ierr)


  ! allocate memory for gather arrays
  ALLOCATE (tri_z(ne))
  ALLOCATE (tri_nz(ne))
  ALLOCATE (tri_sdev(ne))
  ALLOCATE (alltri_z(ne))
  ALLOCATE (alltri_nz(ne))
  ALLOCATE (alltri_sdev(ne))


! initialize tri arrays
	tri_z = zero
	tri_nz = zero
	tri_sdev = zero


  ! 1) gather values
  do i = mystart, mystop
     tri_z(i) = getZee(element(i))
     tri_nz(i) = getNZ(element(i))
     tri_sdev(i) = getSdev(element(i))
  end do



  alltri_z = -999
  alltri_nz = -999
  alltri_sdev = -999
  ! 2) all exchange results
  ! note MPI_IN_PLACE does not seem to be in this MPI implementation!!
  call mpi_allgatherv (tri_nz(mystart),blocklengths(myrank+1), MPI_INTEGER, &
       alltri_nz, blocklengths, displ, MPI_INTEGER, &
       MPI_COMM_WORLD, ierr)
  call mpi_allgatherv (tri_sdev(mystart),blocklengths(myrank+1), MPI_REAL8, &
       alltri_sdev, blocklengths, displ, MPI_REAL8, &
       MPI_COMM_WORLD, ierr)
  call mpi_allgatherv (tri_z(mystart),blocklengths(myrank+1), MPI_REAL8, &
       alltri_z, blocklengths, displ, MPI_REAL8, &
       MPI_COMM_WORLD, ierr)


  ! 3) scatter results back to the triangle objects
  do i = 1, ne
     call setZee(element(i),alltri_z(i),alltri_nz(i),alltri_sdev(i))
  enddo
1111	format (a17, i8, 2f12.4)
1112	format (a17, i8, 2i8)




  ! deallocate memory for gather arrays
  IF (ALLOCATED(tri_z)) DEALLOCATE (tri_z)
  IF (ALLOCATED(tri_nz)) DEALLOCATE (tri_nz)
  IF (ALLOCATED(tri_sdev)) DEALLOCATE (tri_sdev)
  IF (ALLOCATED(alltri_z)) DEALLOCATE (alltri_z)
  IF (ALLOCATED(alltri_nz)) DEALLOCATE (alltri_nz)
  IF (ALLOCATED(alltri_sdev)) DEALLOCATE (alltri_sdev)




END SUBROUTINE exchangeTriangles
! **************************************************************************
! **************************************************************************

                
		!---------------------------------------------------
		! Find the rows and columns of data from the DEM 
		! to use for a given element in the ADCIRC grid
		!---------------------------------------------------

		
		SUBROUTINE FindPatch (n,n_row_n,n_row_s,n_col_w,n_col_e)
			INTEGER,	INTENT (IN) :: n
			INTEGER,    INTENT  (OUT) :: n_col_w,n_col_e,n_row_n,n_row_s
		
           real*8 y_diff_n, y_diff_s, x_diff_w, x_diff_e
           real*8 decimal
           real*8 north, south, west, east
           real*8 row, col

!  Make sure the rectangle falls within the DEM
          west = getWest(element(n))
          east = getEast(element(n))
          south = getSouth(element(n))
          north = getNorth(element(n))


          
          if ((south>=dem_south).and.(north<=dem_north).and.(west>=dem_west).and.(east<=dem_east)) then
!   Calculate the differences from the DEM to the desired point of the rectangle
             x_diff_w=dem_west-west                !xulcorner of DEM - western most point in rectangle
             y_diff_n=dem_north-north              !yulcorner of DEM - northern most point in rectangle
             x_diff_e=dem_west-east                !xulcorner of DEM - eastern most point in rectangle
             y_diff_s=dem_north-south              !yulcorner of DEM - southern most point in rectangle
      
!   Calculate northern most row to use      
             n_row_n=abs(y_diff_n/dxy)
             row=abs(y_diff_n/dxy)
             decimal=row-n_row_n
             if (decimal > half) then
                n_row_n=n_row_n+1
             end if
             if (n_row_n==0) n_row_n=1
             
      
!   Calculate southern most row to use      
             n_row_s=abs(y_diff_s/dxy)
             row=abs(y_diff_s/dxy)
             decimal=row-n_row_s
             if (decimal > half) then
                n_row_s=n_row_s+1
             end if
             if (n_row_s==0) n_row_s=1
            

!   Calculate western-most column to use
             n_col_w=abs(x_diff_w/dxy)
             col=abs(x_diff_w/dxy)
             decimal=col-n_col_w
             if (decimal >= half) then
                n_col_w=n_col_w+1
             end if
             if (n_col_w==0) n_col_w=1

!   Calculate western-most column to use
             n_col_e=abs(x_diff_e/dxy)
             col=abs(x_diff_e/dxy)
             decimal=col-n_col_e
             if (decimal >= half) then
                n_col_e=n_col_e+1
             end if
             if (n_col_e==0) n_col_e=1
          else 
	     n_col_w = 0
	     n_col_e = -1
	     n_row_n = 0
	     n_row_s = -1
	  end if    
	

		END SUBROUTINE FindPatch
		

		!-------------------------------------------------
		! Print an error message, then terminate execution
		!-------------------------------------------------
		SUBROUTINE killJob(message)
			CHARACTER(LEN = *), INTENT(IN) :: message
			WRITE (*,'(a)') message
			STOP "Program terminated by SUBROUTINE killJob"
		END SUBROUTINE killJob

END MODULE DEMClass
