!=================================================================
!=================================================================
!=================================================================
!      =====                                           =====
!      =====             MODULE FileClass              =====
!      =====                                           =====
!=================================================================
!=================================================================
!=================================================================

!===================================================
! This module provides basic file i/o functionality.
!===================================================
MODULE FileClass

	!----------------------------
	! Import debugging capability
	!----------------------------
	USE DebugClass

	!-------------------
	! Import stack class
	!-------------------
	USE StackClass

	!--------------------------------------------
	! Force explicit declaration of all variables
	!--------------------------------------------
	IMPLICIT none

	!-----------------------
	! Set access permissions
	!-----------------------
!	PRIVATE ::
	PUBLIC ::	getUnit, returnUnit, openReadFile, openWriteFile,	&
				closeFile, releaseFile

	!--------------------------------------------------------------
	! Retain values of class instance variables between invocations
	!--------------------------------------------------------------
	SAVE

	!-------------------------
	! Class instance variables
	!-------------------------
	TYPE (intStack), STATIC :: stack
	INTEGER, PARAMETER :: lowUnit = 10, highUnit = 99

	CONTAINS

		!-------------------------------------------------------
		! Purpose:
		!	Obtain a free logical unit from a stack of integers.
		!
		! On input:
		!	n/a
		!
		! On output:
		!	returns -	a unique logical unit for file I/O.
		!-------------------------------------------------------
		FUNCTION getUnit() RESULT(fileUnit)
			INTEGER :: fileUnit
			LOGICAL, SAVE :: firstCall = .TRUE.
			LOGICAL :: busy			! Is the unit already in use?

			IF (firstCall) THEN
				! Create a stack and fill it with logical unit numbers
				CALL new(stack, highUnit - lowUnit + 1)
				CALL fill(stack, highUnit, lowUnit)
				firstCall = .FALSE.
			END IF

			! Find a free logical unit
			busy = .TRUE.
			DO WHILE (busy)
				fileUnit = pop(stack)
				INQUIRE (UNIT=fileUnit, OPENED=busy)
			END DO

			IF (isDebugOn() .AND. getDebugLevel() == INFO) THEN
				WRITE (*, "(a,i3)")										&
					'INFO: Checked out logical unit #', fileUnit
			END IF
		END FUNCTION getUnit

		!--------------------------------------------------
		! Purpose:
		!	Returns a logical unit to the stack.
		!
		! On input:
		!	fileUnit -	Logical unit to be checked back in.
		!
		! On output:
		!	n/a
		!--------------------------------------------------
		SUBROUTINE returnUnit(fileUnit)
			INTEGER, INTENT(IN) :: fileUnit
			LOGICAL :: busy			! Is the unit still in use?

			INQUIRE (UNIT=fileUnit, OPENED=busy)

			IF (.NOT. busy) THEN
				! Push logical unit back onto integer stack
				CALL push(stack, fileUnit)

				IF (isDebugOn() .AND. getDebugLevel() == INFO) THEN
					WRITE	(*, "(a,i3)")								&
						'INFO: Returned    logical unit #',	 fileUnit
				END IF
			END IF
		END SUBROUTINE returnUnit

		!----------------------------------------------
		! Purpose:
		! 	Open a file for reading formatted data.
		!
		! On input:
		!	fileUnit -	Logical unit for input file
		!	fileName -	Input file name
		!
		! On output:
		!	returns -	.TRUE.  if open was successful;
		!				.FALSE. if open failed.
		!----------------------------------------------
		LOGICAL FUNCTION openReadFile(fileUnit, fileName)
			INTEGER,		INTENT(IN)	:: fileUnit
			CHARACTER (*),	INTENT(IN)	:: fileName
			INTEGER						:: anyErrors	! Status: 0 for success, else failure

			OPEN (UNIT=fileUnit, FILE=TRIM(ADJUSTL(fileName)),			&
				  STATUS='OLD', ACTION='READ', IOSTAT=anyErrors)

			openReadFile = (anyErrors == 0)
		END FUNCTION openReadFile

		!----------------------------------------------
		! Purpose:
		! 	Open a file for writing formatted data.
		!
		! On input:
		!	fileUnit -	Logical unit for input file
		!	fileName -	Input file name
		!
		! On output:
		!	returns -	.TRUE.  if open was successful;
		!				.FALSE. if open failed.
		!----------------------------------------------
		LOGICAL FUNCTION openWriteFile(fileUnit, fileName, recordLength)
			INTEGER,		INTENT(IN)				:: fileUnit
			CHARACTER (*),	INTENT(IN)				:: fileName
			INTEGER,		INTENT(IN), OPTIONAL	:: recordLength
			INTEGER									:: anyErrors	! Status: 0 for success, else failure

			IF ( PRESENT(recordLength) ) THEN
				OPEN (UNIT=fileUnit, FILE=TRIM(ADJUSTL(fileName)),		&
					RECL=recordLength, STATUS='UNKNOWN',				&
					ACTION='WRITE', IOSTAT=anyErrors)
			ELSE
				OPEN (UNIT=fileUnit, FILE=TRIM(ADJUSTL(fileName)),		&
					STATUS='UNKNOWN', ACTION='WRITE', IOSTAT=anyErrors)
			END IF

			openWriteFile = (anyErrors == 0)
		END FUNCTION openWriteFile

		!------------------------------------------
		! Purpose:
		!	Close a file.
		!
		! On input:
		!	fileUnit -	Logical unit for input file
		!
		! On output:
		!	File is closed.
		!------------------------------------------
		SUBROUTINE closeFile(fileUnit)
			INTEGER, INTENT(IN) :: fileUnit
			LOGICAL :: isOpen	! Is the file open?

			INQUIRE (UNIT=fileUnit, OPENED=isOpen)
			IF (isOpen) THEN
				CLOSE (UNIT=fileUnit)
			END IF
		END SUBROUTINE closeFile

		!-----------------------------------------------------
		! Purpose:
		!	Both close a file and release its logical unit.
		!
		! On input:
		!	fileUnit -	Logical unit for input file
		!
		! On output:
		!	File is closed and logical unit returned to stack.
		!-----------------------------------------------------
		SUBROUTINE releaseFile(fileUnit)
			INTEGER, INTENT(IN) :: fileUnit
			CALL closeFile (fileUnit)
			CALL returnUnit(fileUnit)
		END SUBROUTINE releaseFile

		!------------------------------------
		! Print a diagnostic file I/O message
		!------------------------------------
		SUBROUTINE checkFileStatus(anyErrors)
			INTEGER, INTENT(IN)	:: anyErrors			! File I/O status
			IF (anyErrors < 0) THEN						! We've hit end-of-file
				WRITE (*,'(a)') "End-of-file on read"
			ELSE IF (anyErrors > 0) THEN				! User error
				WRITE (*,'(a)') "User file read error!"
			ELSE										! No errors
!				WRITE (*,'(a)') "Data read successfully"
			END IF
		END SUBROUTINE checkFileStatus

END MODULE FileClass
