!=================================================================
!=================================================================
!=================================================================
!      =====                                           =====
!      =====             MODULE PointClass             =====
!      =====                                           =====
!=================================================================
!=================================================================
!=================================================================

!==========================================================
! This module defines a Point object and its functionality.
!==========================================================
MODULE PointClass

	!------------------------
	! Import global constants
	!------------------------
	USE ConstantsClass, ONLY: zero

	!--------------------------------------------
	! Force explicit declaration of all variables
	!--------------------------------------------
	IMPLICIT none

	!-----------------------
	! Set access permissions
	!-----------------------
	PRIVATE	::	newPointR, newPointD
	PUBLIC	::	newPoint, deletePoint, printPoint, writePoint,				&
					setID,getID, setX,setY,setZ, getX,getY,getZ, convertZ,	&
					setNeighbors, getNeighboringPoints, getNeighboringShapes

	!-----------------------------------
	! Define the Point derived data type
	!-----------------------------------
	TYPE Point
		PRIVATE
		INTEGER	(4) :: id													! ID number
		REAL		(8) :: x, y, z												! Coordinates
		INTEGER	(4) :: nnp													! Number of neighoring points
		INTEGER	(4) :: nns													! Number of neighoring shapes
		INTEGER	(4),	ALLOCATABLE,	DIMENSION(:) :: nPoints		! List   of neighoring points
!		INTEGER	(4),	ALLOCATABLE,	DIMENSION(:) :: nShapes		! List   of neighoring shapes
		INTEGER	(4),	DIMENSION(16) :: nShapes		! List   of neighoring shapes
	END TYPE Point

	!--------------------------------------------------------------
	! Retain values of class instance variables between invocations
	!--------------------------------------------------------------
	SAVE

	!-------------------------
	! Class instance variables
	!-------------------------

	!-----------------------------
	! Define constructor interface
	!-----------------------------
	INTERFACE newPoint
		MODULE PROCEDURE newPointR, newPointD
	END INTERFACE

	!----------------------------
	! Define destructor interface
	!----------------------------
	INTERFACE delete
		MODULE PROCEDURE deletePoint
	END INTERFACE

	!-----------------------
	! Define print interface
	!-----------------------
	INTERFACE print
		MODULE PROCEDURE printPoint
	END INTERFACE

	CONTAINS

		!-------------
		! Constructors
		!-------------
		TYPE (Point) FUNCTION newPointR(id, x, y, z) RESULT(this)
			INTEGER	(4), INTENT (IN) :: id
			REAL		(4), INTENT (IN) :: x, y, z
			this%id = id
			this%x = DBLE(x)
			this%y = DBLE(y)
			this%z = DBLE(z)
		END FUNCTION newPointR

		TYPE (Point) FUNCTION newPointD(id, x, y, z) RESULT(this)
			INTEGER	(4), INTENT (IN) :: id
			REAL		(8), INTENT (IN) :: x, y, z
			this%id = id
			this%x = x
			this%y = y
			this%z = z
		END FUNCTION newPointD

		!-----------
		! Destructor
		!-----------
		SUBROUTINE deletePoint(this)
			TYPE (Point), INTENT (IN OUT) :: this
			this%id	= 0
			this%x	= zero
			this%y	= zero
			this%z	= zero
			this%nnp	= 0
			this%nns	= 0
!			IF (ALLOCATED(this%nShapes))  DEALLOCATE(this%nShapes)
!			IF (ALLOCATED(this%nPoints))  DEALLOCATE(this%nPoints)
		END SUBROUTINE deletePoint

		!---------------------------
		! Setter (mutator) functions
		!---------------------------
		SUBROUTINE setID(this, id)
			TYPE (Point),	INTENT(IN OUT)	:: this
			INTEGER (4),	INTENT(IN)		:: id
			this%id = id
		END SUBROUTINE setID

		SUBROUTINE setX(this, x)
			TYPE (Point),	INTENT(IN OUT)	:: this
			REAL(8),			INTENT(IN)		:: x
			this%x = x
		END SUBROUTINE setX

		SUBROUTINE setY(this, y)
			TYPE (Point),	INTENT(IN OUT)	:: this
			REAL(8),			INTENT(IN)		:: y
			this%y = y
		END SUBROUTINE setY

		SUBROUTINE setZ(this, z)
			TYPE (Point),	INTENT(IN OUT)	:: this
			REAL(8),			INTENT(IN)		:: z
			this%z = z
		END SUBROUTINE setZ

		SUBROUTINE convertZ(this, factor)
			TYPE (Point),	INTENT(IN OUT)	:: this
			REAL(8),			INTENT(IN)		:: factor
			this%z = this%z * factor
		END SUBROUTINE convertZ

		SUBROUTINE setNeighbors(this, nnp,nns, nPoints,nShapes)
			TYPE (Point),	INTENT (IN OUT)										:: this
			INTEGER	(4),	INTENT (IN)												:: nnp
			INTEGER	(4),	INTENT (IN)												:: nns
			INTEGER	(4),	INTENT (IN), DIMENSION(nnp)						:: nPoints
			INTEGER	(4),	INTENT (IN), DIMENSION(nns)						:: nShapes
!			INTEGER	(4),	INTENT (IN OUT), ALLOCATABLE,	DIMENSION(:)	:: nPoints
!			INTEGER	(4),	INTENT (IN OUT), ALLOCATABLE,	DIMENSION(:)	:: nShapes
!			INTEGER	(4),	INTENT (IN OUT), DIMENSION(nnp)	:: nPoints
!			INTEGER	(4),	INTENT (IN OUT), DIMENSION(nns)	:: nShapes
			this%nnp = nnp
			this%nns = nns
!			IF (.NOT. ALLOCATED(this%nPoints)) ALLOCATE(this%nPoints(nnp))
			this%nPoints = nPoints
!			IF (.NOT. ALLOCATED(this%nShapes)) ALLOCATE(this%nShapes(nns))
			this%nShapes = nShapes
		END SUBROUTINE setNeighbors

		!---------------------------
		! Getter (accessor) functions
		!---------------------------
		INTEGER(4) FUNCTION getID(this)
			TYPE (Point), INTENT(IN) :: this
			getID = this%id
		END FUNCTION getID

		REAL(8) FUNCTION getX(this)
			TYPE (Point), INTENT(IN) :: this
			getX = this%x
		END FUNCTION getX

		REAL(8) FUNCTION getY(this)
			TYPE (Point), INTENT(IN) :: this
			getY = this%y
		END FUNCTION getY

		REAL(8) FUNCTION getZ(this)
			TYPE (Point), INTENT(IN) :: this
			getZ = this%z
		END FUNCTION getZ

		INTEGER(4) FUNCTION getNumberOfNeighboringPoints(this)
			TYPE (Point), INTENT(IN) :: this
			getNumberOfNeighboringPoints = this%nnp
		END FUNCTION getNumberOfNeighboringPoints

		SUBROUTINE getNeighboringPoints(this, nnp, nPoints)
			TYPE (Point),	INTENT (IN)							:: this
			INTEGER	(4),	INTENT (IN OUT)						:: nnp
			INTEGER	(4),	INTENT (OUT), DIMENSION(nnp)	:: nPoints
			nnp		= MIN(nnp, this%nnp)
			nPoints	= this%nPoints(1:nnp)
		END SUBROUTINE getNeighboringPoints

		INTEGER(4) FUNCTION getNumberOfNeighboringShapes(this)
			TYPE (Point), INTENT(IN) :: this
			getNumberOfNeighboringShapes = this%nns
		END FUNCTION getNumberOfNeighboringShapes

		SUBROUTINE getNeighboringShapes(this, nns, horse)
			TYPE (Point),	INTENT (IN)							:: this
			INTEGER	(4),	INTENT (IN OUT)						:: nns
			INTEGER	(4),	INTENT (OUT), DIMENSION(16)	:: horse
!			nns		= MIN(nns, this%nns)
!            write(*,*) this%nShapes
 		    horse	= this%nShapes
		END SUBROUTINE getNeighboringShapes

		!-------------------
		! Print Point object
		!-------------------
		SUBROUTINE printPoint(this)
			TYPE (Point), INTENT (IN) :: this
			WRITE (*,'(i7,". (",2(f12.3,","),f12.3,")")')						&
				this%id, this%x,this%y,this%z
		END SUBROUTINE printPoint

		!-------------------
		! Write Point object
		!-------------------
		SUBROUTINE writePoint(this, unit)
			TYPE (Point),	INTENT (IN) 				:: this
			INTEGER,			INTENT (IN), OPTIONAL	:: unit
			IF (PRESENT(unit)) THEN
				WRITE (unit, '(i10, 2(1x,f18.10), 1x,f15.10)')					&
					this%id, this%x,this%y,this%z
			ELSE
				WRITE (*,    '(i10, 2(1x,f18.10), 1x,f15.10)')					&
					this%id, this%x,this%y,this%z
			END IF
		END SUBROUTINE writePoint

		!--------------------------
		! Print a Point's neighbors
		!--------------------------
		SUBROUTINE printNeighbors(this)
			TYPE (Point), INTENT (IN) :: this
			WRITE (*,'(/,i6," neighboring points for point #", i6, ":")')	&
					this%nnp, this%id
			WRITE (*,'(10(1x,i7))') this%nPoints
			WRITE (*,'(i6," neighboring shapes for point #", i6, ":")')		&
					this%nns, this%id
			WRITE (*,'(10(1x,i7))') this%nShapes
		END SUBROUTINE printNeighbors

END MODULE PointClass
