!=================================================================
!=================================================================
!=================================================================
!      =====                                           =====
!      =====             MODULE StackClass             =====
!      =====                                           =====
!=================================================================
!=================================================================
!=================================================================

!=====================================================================
! This MODULE defines the integer Stack class, as described in the C++
! Primer, by Stanley B. Lippman [Addison-Wesley, Reading, MA, 1991],
! pp. 73-77.
!=====================================================================
MODULE StackClass

	!--------------------------------------------
	! Force explicit declaration of all variables
	!--------------------------------------------
	IMPLICIT none

	!-----------
	! Set access
	!-----------
	PRIVATE
	PUBLIC ::	intStack
	PUBLIC ::	new, destroy, fill, isEmpty, isFull, push, pop, &
				getSize, getTop, getBot, getValueAt, printStack

	INTEGER, PARAMETER :: BOS = 0			! Bottom of stack
	INTEGER, PARAMETER :: STACK_SIZE = 32	! Default stack size

	!--------------------------------------
	! Define the intStack derived data type
	!--------------------------------------
	TYPE intStack
		PRIVATE
		INTEGER :: size, top
		INTEGER, DIMENSION (:), POINTER :: array
	END TYPE intStack

	!-----------------------------
	! Define constructor interface
	!-----------------------------
	INTERFACE new
		MODULE PROCEDURE newStack
	END INTERFACE

	!----------------------------
	! Define destructor interface
	!----------------------------
	INTERFACE destroy
		MODULE PROCEDURE destroyStack
	END INTERFACE

	CONTAINS

		!------------
		! Constructor
		!------------
		SUBROUTINE newStack(this, sz)
			TYPE (intStack) :: this

			INTEGER, INTENT(IN), OPTIONAL :: sz
			INTEGER :: size

			IF (PRESENT(sz)) THEN
				size = sz
			ELSE
				size = STACK_SIZE
			END IF

			ALLOCATE(this%array(size))

			IF (ASSOCIATED(this%array)) THEN
				this%top = BOS
				this%size = size
			END IF
		END SUBROUTINE newStack

		!-----------
		! Destructor
		!-----------
		SUBROUTINE destroyStack(this)
			TYPE (intStack) :: this
			IF (ASSOCIATED(this%array)) THEN
				DEALLOCATE(this%array)
			END IF
		END SUBROUTINE destroyStack

		!---------------
		! Fill the stack
		!---------------
		SUBROUTINE fill(this, low, high)
			TYPE (intStack) :: this
			INTEGER, INTENT(IN), OPTIONAL :: low, high
			INTEGER :: n, lowValue, highValue

			IF (PRESENT(low)) THEN
				lowValue = low
			ELSE
				lowValue = 1
			END IF

			IF (PRESENT(high)) THEN
				highValue = high
			ELSE
				highValue = STACK_SIZE
			END IF

			DO n = lowValue, highValue, SIGN(1, highValue-lowValue)
				CALL push(this, n)
			END DO
		END SUBROUTINE fill

		!--------------------
		! Is the stack empty?
		!--------------------
		LOGICAL FUNCTION isEmpty(this)
			TYPE (intStack) :: this
			isEmpty = this%top .EQ. BOS
		END FUNCTION isEmpty

		!-------------------
		! Is the stack full?
		!-------------------
		LOGICAL FUNCTION isFull(this)
			TYPE (intStack) :: this
			isFull = this%top .EQ. this%size
		END FUNCTION isFull

		!------------------------
		! Push a value onto stack
		!------------------------
		SUBROUTINE push(this, value)
			TYPE (intStack) :: this
			INTEGER, INTENT(IN) :: value
			IF (isFull(this)) THEN
				WRITE (*,*) 'ERROR: stack overflow'
				STOP 2
			ELSE
				this%top = this%top + 1
				this%array(this%top) = value
			END IF
		END SUBROUTINE push

		!--------------------------------------
		! Pop the topmost element off the stack
		!--------------------------------------
		INTEGER FUNCTION pop(this)
			TYPE (intStack) :: this
			IF (isEmpty(this)) THEN
				WRITE (*,*) 'ERROR: stack underflow'
				STOP 1
			ELSE
				pop = this%array(this%top)
				this%top = this%top - 1
			END IF
		END FUNCTION pop

		!--------------------------
		! Get the size of the stack
		!--------------------------
		INTEGER FUNCTION getSize(this)
			TYPE (intStack) :: this
			getSize = this%size
		END FUNCTION getSize

		!--------------------
		! Get the top element
		!--------------------
		INTEGER FUNCTION getTop(this)
			TYPE (intStack) :: this
			getTop = this%array(this%top)
		END FUNCTION getTop

		!-----------------------
		! Get the bottom element
		!-----------------------
		INTEGER FUNCTION getBot(this)
			TYPE (intStack) :: this
			getBot = this%array(1)
		END FUNCTION getBot

		!----------------------------------
		! Get a value located at position n
		!----------------------------------
		INTEGER FUNCTION getValueAt(this, n)
			TYPE (intStack) :: this
			INTEGER, INTENT(IN) :: n
			IF (n < 1) THEN
				WRITE (*,*) 'ERROR: stack underflow'
				STOP 1
			ELSE IF (n > this%top) THEN
				WRITE (*,*) 'ERROR: stack overflow'
				STOP 2
			ELSE
				getValueAt = this%array(n)
			END IF
		END FUNCTION getValueAt

		!------------------------------
		! Print all elements in a stack
		!------------------------------
		SUBROUTINE printStack(this, fileUnit)
			TYPE (intStack), INTENT(IN) :: this
			INTEGER, INTENT(IN) :: fileUnit
			WRITE (UNIT=fileUnit, FMT='(A)') '['
			WRITE (UNIT=fileUnit, FMT='((9(I5, ", "), I5))') this%array
			WRITE (UNIT=fileUnit, FMT='(A)') ']'
		END SUBROUTINE printStack

END MODULE StackClass
