!=================================================================
!=================================================================
!=================================================================
!      =====                                           =====
!      =====             PROGRAM TestDEM             =====
!      =====                                           =====
!=================================================================
!=================================================================
!=================================================================

!=================================================================
! Run some unit tests that process the NC Floodplain Mapping
! System's LIDAR DEM 20' topo files.
!
! Compile:
!    make
!
! Compile and run:
!    make run
!
! Clean up:
!    make clean
!    make realclean
!
! Revision history:
!    Date         Programmer                 Description of change
!    ----         ----------                 ---------------------
!    03/20/07     Craig  Mattocks, UNC-IMS   Wrote original code
!    10/04/07     Crystal Fulcher, UNC-IMS   Changes to include
!                                            DEMClass instead of
!                                            LidarClass to use with
!                                            RENCI DEM
!=================================================================
PROGRAM TestDEM

	!-----------------------
	! Import CPU timer class
	!-----------------------
	USE TimerClass

	!-------------------
	! Import Point class
	!-------------------
	USE PointClass

	!-----------------------------------
	! Import lidar data processing class
	!-----------------------------------
	USE DEMClass

	!--------------------------------------------
	! Force explicit declaration of all variables
	!--------------------------------------------
	IMPLICIT none

	!----------------------------
	! Execution control variables
	!----------------------------
	LOGICAL, PARAMETER :: stopNow = .TRUE.

	!---------------------
	! CPU timing variables
	!---------------------
	REAL    :: cputime = 0.				! Net execution time
	INTEGER :: minutes = 0				! Minutes required for execution and
	REAL    :: seconds = 0.				! seconds required for execution

	!----------------
	! Local variables
	!----------------
	INTEGER :: n							! Loop index
	INTEGER :: DEMFileUnit
	INTEGER :: nprocs, myrank, ierr
	REAL*8 :: starttime, stoptime, walltime


! MPI include file
	include "mpif.h"


! Initialize MPI
	call MPI_Init(ierr)
	call MPI_COMM_Size(MPI_COMM_WORLD, nprocs, ierr)
	call MPI_COMM_Rank(MPI_COMM_WORLD, myrank, ierr)


	!---------------------------------
	! User requested early termination
	!---------------------------------
!	IF (stopNow) GO TO 99

	!---------------------------------------
	! Run some unit tests on the lidar class
	!---------------------------------------
	CALL tic									! Start CPU time "stopwatch" ...
	starttime = MPI_Wtime()

	!------------------
	! Obtain user input
	!------------------
	CALL userInput(inputGridFile, outputGridFile, asciiDEM, dir_accessDEM)

	!-----------------------------------------
	! Dump some LiDAR data tiles in GMT format
	!-----------------------------------------
!	DO n = 1, ntf
!		IF (readLidar(TRIM(ADJUSTL(lidarTileDir))//tileFile(n))) THEN
!			WRITE(*,*) "Writing file ", tileFile(n)(1:24)//".gmt"
!			CALL writeLidar(tileFile(n)(1:24)//".gmt")
!		END IF
!	END DO

	!---------------------
	! Read input grid file
	!---------------------
	CALL readGrid(inputGridFile,outputGridFile)
	

	CALL readDEM(asciiDEM,DEMFileUnit)

	!---------------------------------------------------
	! Compute the average (or envelope) terrain height
	! from LiDAR 20' DEM data for all ADCIRC grid nodes.
	!---------------------------------------------------
	CALL computeTerrain(DEMFileUnit)
	
	!-----------------------
	! Write output grid file
	!-----------------------
!	CALL writeGrid(outputGridFile)

	!------------
	! Free memory
	!------------
 	CALL deleteGrid()

	!--------------------------------------
	! Program has reached termination point
	!--------------------------------------
99	CONTINUE



! Exit MPI
	stoptime = MPI_Wtime()
	call MPI_Barrier(MPI_COMM_WORLD, ierr)
	call MPI_Finalize (ierr)
	

	! ---------------------------------------------
	! Calculate the CPU time required for execution
	! ---------------------------------------------
	cputime = toc()				! ... stop CPU time "stopwatch"
	minutes = INT (cputime / 60.)
	seconds = AMOD(cputime , 60.)

!	WRITE	(*,"(53('-'))")
	WRITE	(*,"('rank= ',i4,' Execution required ',i3,' minutes and ',f9.6,' seconds.')")	&
   		myrank, minutes, seconds
!	WRITE	(*,"(53('-'))")
	write (*,1111) "rank= ",myrank," cputime = ",cputime, " seconds"
	walltime = stoptime-starttime
	write (*,1111) "rank= ",myrank," walltime = ",walltime, " seconds"
1111	format (a6,i4,a12,f15.0,a8)

	STOP

END PROGRAM TestDEM
