!=================================================================
!=================================================================
!=================================================================
!      =====                                           =====
!      =====             PROGRAM TestLidar             =====
!      =====                                           =====
!=================================================================
!=================================================================
!=================================================================

!=================================================================
! Run some unit tests that process the NC Floodplain Mapping
! System's LIDAR DEM 20' topo files.
!
! Compile:
!    make
!
! Compile and run:
!    make run
!
! Clean up:
!    make clean
!    make realclean
!
! Revision history:
!    Date         Programmer                 Description of change
!    ----         ----------                 ---------------------
!    03/20/07     Craig  Mattocks, UNC-IMS   Wrote original code
!=================================================================
PROGRAM TestLidar

	!-----------------------
	! Import CPU timer class
	!-----------------------
	USE TimerClass

	!-------------------
	! Import Point class
	!-------------------
	USE PointClass

	!-----------------------------------
	! Import lidar data processing class
	!-----------------------------------
	USE LidarClass

	!--------------------------------------------
	! Force explicit declaration of all variables
	!--------------------------------------------
	IMPLICIT none

	!----------------------------
	! Execution control variables
	!----------------------------
	LOGICAL, PARAMETER :: stopNow = .TRUE.

	!---------------------
	! CPU timing variables
	!---------------------
	REAL    :: cputime = 0.				! Net execution time
	INTEGER :: minutes = 0				! Minutes required for execution and
	REAL    :: seconds = 0.				! seconds required for execution

	!----------------
	! Local variables
	!----------------
	INTEGER :: n							! Loop index

	!-----------------
	! LiDAR tile files
	!-----------------
!	INTEGER, PARAMETER :: ntf = 42
!	CHARACTER*28	::	tileFile(ntf) =	(/											&
!													"nc_lidar_dem20_x250_y028.dat",	&
!													"nc_lidar_dem20_x250_y029.dat",	&
!													"nc_lidar_dem20_x250_y030.dat",	&
!													"nc_lidar_dem20_x250_y031.dat",	&
!													"nc_lidar_dem20_x250_y032.dat",	&
!													"nc_lidar_dem20_x250_y033.dat",	&
!													"nc_lidar_dem20_x250_y034.dat",	&
!													"nc_lidar_dem20_x250_y035.dat",	&
!													"nc_lidar_dem20_x250_y036.dat",	&
!													"nc_lidar_dem20_x251_y029.dat",	&
!													"nc_lidar_dem20_x251_y030.dat",	&
!													"nc_lidar_dem20_x251_y031.dat",	&
!													"nc_lidar_dem20_x251_y032.dat",	&
!													"nc_lidar_dem20_x251_y033.dat",	&
!													"nc_lidar_dem20_x251_y034.dat",	&
!													"nc_lidar_dem20_x251_y035.dat",	&
!													"nc_lidar_dem20_x251_y036.dat",	&
!													"nc_lidar_dem20_x252_y030.dat",	&
!													"nc_lidar_dem20_x252_y031.dat",	&
!													"nc_lidar_dem20_x252_y032.dat",	&
!													"nc_lidar_dem20_x252_y033.dat",	&
!													"nc_lidar_dem20_x252_y034.dat",	&
!													"nc_lidar_dem20_x252_y035.dat",	&
!													"nc_lidar_dem20_x252_y036.dat",	&
!													"nc_lidar_dem20_x253_y030.dat",	&
!													"nc_lidar_dem20_x253_y031.dat",	&
!													"nc_lidar_dem20_x253_y032.dat",	&
!													"nc_lidar_dem20_x253_y033.dat",	&
!													"nc_lidar_dem20_x253_y034.dat",	&
!													"nc_lidar_dem20_x253_y035.dat",	&
!													"nc_lidar_dem20_x253_y036.dat",	&
!													"nc_lidar_dem20_x254_y031.dat",	&
!													"nc_lidar_dem20_x254_y032.dat",	&
!													"nc_lidar_dem20_x254_y033.dat",	&
!													"nc_lidar_dem20_x254_y034.dat",	&
!													"nc_lidar_dem20_x254_y035.dat",	&
!													"nc_lidar_dem20_x254_y036.dat",	&
!													"nc_lidar_dem20_x255_y032.dat",	&
!													"nc_lidar_dem20_x255_y033.dat",	&
!													"nc_lidar_dem20_x255_y034.dat",	&
!													"nc_lidar_dem20_x255_y035.dat",	&
!													"nc_lidar_dem20_x255_y036.dat"	&
!													/)
!

	!---------------------------------
	! User requested early termination
	!---------------------------------
!	IF (stopNow) GO TO 99

	!---------------------------------------
	! Run some unit tests on the lidar class
	!---------------------------------------
	CALL tic									! Start CPU time "stopwatch" ...

	!------------------
	! Obtain user input
	!------------------
	CALL userInput(inputGridFile, outputGridFile, lidarTileDir)

	!-----------------------------------------
	! Dump some LiDAR data tiles in GMT format
	!-----------------------------------------
!	DO n = 1, ntf
!		IF (readLidar(TRIM(ADJUSTL(lidarTileDir))//tileFile(n))) THEN
!			WRITE(*,*) "Writing file ", tileFile(n)(1:24)//".gmt"
!			CALL writeLidar(tileFile(n)(1:24)//".gmt")
!		END IF
!	END DO

	!---------------------
	! Read input grid file
	!---------------------
	CALL readGrid(inputGridFile)

	!---------------------------------------------------
	! Compute the average (or envelope) terrain height
	! from LiDAR 20' DEM data for all ADCIRC grid nodes.
	!---------------------------------------------------
	CALL computeTerrain()

	!-----------------------
	! Write output grid file
	!-----------------------
	CALL writeGrid(outputGridFile)

	!------------
	! Free memory
	!------------
 	CALL deleteGrid()

	!--------------------------------------
	! Program has reached termination point
	!--------------------------------------
99	CONTINUE

	! ---------------------------------------------
	! Calculate the CPU time required for execution
	! ---------------------------------------------
	cputime = toc()				! ... stop CPU time "stopwatch"
	minutes = INT (cputime / 60.)
	seconds = AMOD(cputime , 60.)

	WRITE	(*,"(53('-'))")
	WRITE	(*,"('Execution required ',i3,' minutes and ',f9.6,' seconds.')")	&
   		minutes, seconds
	WRITE	(*,"(53('-'))")

	STOP

END PROGRAM TestLidar
