!=================================================================
!=================================================================
!=================================================================
!      =====                                           =====
!      =====           MODULE TriangleClass            =====
!      =====                                           =====
!=================================================================
!=================================================================
!=================================================================

!=============================================================
! This module defines a Triangle object and its functionality.
!=============================================================
MODULE TriangleClass

	!---------------------------------
	! Import selected global constants
	!---------------------------------
	USE ConstantsClass, ONLY: zero, half, small, missingDouble

	!-------------------
	! Import Point class
	!-------------------
	USE PointClass

	!--------------------------------------------
	! Force explicit declaration of all variables
	!--------------------------------------------
	IMPLICIT none

	!-----------------------
	! Set access permissions
	!-----------------------
	PRIVATE	::	calcArea, calcExtent
	PUBLIC	::	newTriangle, deleteTriangle,										&
					printTriangle, writeTriangle,										&
					setNumber,getNumber, getArea,										&
					getNorth, getSouth, getEast, getWest,							&
					getP1, getP2, getP3,													&
					addZ, calcZ, getNZ, getZee, getSdev, setZee

	!--------------------------------------
	! Define the Triangle derived data type
	!--------------------------------------
	TYPE Triangle
		PRIVATE
		INTEGER (4)		:: number
		TYPE (Point)	:: p1, p2, p3
		REAL (8)			:: area
		REAL (8)			:: north, south, east, west
		INTEGER (8)		:: nz
		REAL (8)			:: z
		REAL (8)			:: sdev
	END TYPE Triangle

	!-----------------------------
	! Define constructor interface
	!-----------------------------
	INTERFACE new
		MODULE PROCEDURE newTriangle
	END INTERFACE

	!----------------------------
	! Define destructor interface
	!----------------------------
	INTERFACE delete
		MODULE PROCEDURE deleteTriangle
	END INTERFACE

	!-----------------------
	! Define print interface
	!-----------------------
	INTERFACE print
		MODULE PROCEDURE printTriangle
	END INTERFACE

	CONTAINS

		!------------
		! Constructor
		!------------
		TYPE (Triangle) FUNCTION newTriangle(n, p1,p2,p3) RESULT(this)
			INTEGER (4),	INTENT (IN) :: n
			TYPE (Point),	INTENT (IN) :: p1, p2, p3
			this%number = n
			this%p1 = p1
			this%p2 = p2
			this%p3 = p3
			CALL calcExtent(this)
			this%area = calcArea(this)
			this%nz		= 0
			this%z		= zero
			this%sdev	= zero
		END FUNCTION newTriangle

		!-----------------------
		! Delete Triangle object
		!-----------------------
		SUBROUTINE deleteTriangle(this)
			TYPE (Triangle), INTENT (IN OUT) :: this
			this%number		= 0
			CALL deletePoint(this%p1)
			CALL deletePoint(this%p2)
			CALL deletePoint(this%p3)
			this%area		= zero
			this%north		= zero
			this%south		= zero
			this%east		= zero
			this%west		= zero
			this%nz			= 0
			this%z			= zero
			this%sdev		= zero
		END SUBROUTINE deleteTriangle

		!----------------------------
		! Getter (accessor) functions
		!----------------------------
		TYPE (Point) FUNCTION getP1(this)
			TYPE (Triangle), INTENT(IN) :: this
			getP1 = this%p1
		END FUNCTION getP1

		TYPE (Point) FUNCTION getP2(this)
			TYPE (Triangle), INTENT(IN) :: this
			getP2 = this%p2
		END FUNCTION getP2

		TYPE (Point) FUNCTION getP3(this)
			TYPE (Triangle), INTENT(IN) :: this
			getP3 = this%p3
		END FUNCTION getP3

		INTEGER(4) FUNCTION getNumber(this)
			TYPE (Triangle), INTENT(IN) :: this
			getNumber = this%number
		END FUNCTION getNumber

		REAL(8) FUNCTION getArea(this)
			TYPE (Triangle), INTENT(IN) :: this
			getArea = this%area
		END FUNCTION getArea

		REAL(8) FUNCTION getNorth(this)
			TYPE (Triangle), INTENT(IN) :: this
			getNorth = this%north
		END FUNCTION getNorth

		REAL(8) FUNCTION getSouth(this)
			TYPE (Triangle), INTENT(IN) :: this
			getSouth = this%south
		END FUNCTION getSouth

		REAL(8) FUNCTION getEast(this)
			TYPE (Triangle), INTENT(IN) :: this
			getEast = this%east
		END FUNCTION getEast

		REAL(8) FUNCTION getWest(this)
			TYPE (Triangle), INTENT(IN) :: this
			getWest = this%west
		END FUNCTION getWest

		INTEGER(4) FUNCTION getNZ(this)
			TYPE (Triangle), INTENT(IN) :: this
			getNZ = this%nz
		END FUNCTION getNZ

		REAL(8) FUNCTION getZee(this)
			TYPE (Triangle), INTENT(IN) :: this
			getZee = this%z
		END FUNCTION getZee

		REAL(8) FUNCTION getSdev(this)
			TYPE (Triangle), INTENT(IN) :: this
			getSdev = this%sdev
		END FUNCTION getSdev

		!---------------------------
		! Setter (mutator) functions
		!---------------------------
		SUBROUTINE setNumber(this, n)
			TYPE (Triangle), INTENT(IN OUT) :: this
			INTEGER (4), INTENT (IN) :: n
			this%number = n
		END SUBROUTINE setNumber

		!-------------------------------------------------------------
		! Calculate the area of a Triangle with vertices
		! P1(x1,y1), P2(x2,y2), and P3(x3,y3) using the determinant:
		!
		! A = abs(1/2 * |x1  y1  1|)
		!               |x2  y2  1|
		!               |x3  y3  1|
		!
		! A = abs(1/2 * [x1 * |y2  1| - y1 * |x2  1|  + 1 * |x2 y2|])
		!                     |y3  1|        |x3  1|        |x3 y3|
		!
		! A = abs(1/2 * [x1*(y2-y3) - y1*(x2-x3) + (x2*y3-y2*x3)])
		!   = abs(1/2 * [(x3-x2)*y1 + (x1-x3)*y2 + (x2-x1)*y3])
		!   = abs(1/2 * [(x1-x3)*(y2-y3) + (x3-x2)*(y1-y3)])
		!
		! On input:
		!	n/a
		!
		! On output:
		!	returns -	the area of a triangle.
		!-------------------------------------------------------------
		REAL(8) FUNCTION calcArea(this)
			TYPE (Triangle), INTENT (IN)	:: this
			REAL (8)								:: x1, y1
			REAL (8)								:: x2, y2
			REAL (8)								:: x3, y3

			x1 = getX(this%p1) - this%west
			x2 = getX(this%p2) - this%west
			x3 = getX(this%p3) - this%west

			y1 = getY(this%p1) - this%south
			y2 = getY(this%p2) - this%south
			y3 = getY(this%p3) - this%south

			calcArea =	ABS( half * ((x1-x3)*(y2-y3) + (x3-x2)*(y1-y3)) )
		END FUNCTION calcArea

		!---------------------------------------------------------------
		! Calculate the rectangular extent of a triangle.
		!
		! On input:
		!	A Triangle
		!
		! On output:
		!	Finds the north, south, east and west edges of the rectangle
		!	that contains the triangle.
		!---------------------------------------------------------------
		SUBROUTINE calcExtent(this)
			TYPE (Triangle), INTENT (IN OUT) :: this
			this%north = MAX(getY(this%p1),getY(this%p2),getY(this%p3))
			this%south = MIN(getY(this%p1),getY(this%p2),getY(this%p3))
			this%east  = MAX(getX(this%p1),getX(this%p2),getX(this%p3))
			this%west  = MIN(getX(this%p1),getX(this%p2),getX(this%p3))
		END SUBROUTINE calcExtent

		!---------------------------------------------------------------
		! Is a Point near (within the rectangular extent of) a Triangle?
		!
		! On input:
		!	this		-	Triangle
		!	p			-	Point to test
		!
		! On output:
		!	returns	-	.TRUE.	if Point is near Triangle
		!					.FALSE.	otherwise
		!---------------------------------------------------------------
		LOGICAL FUNCTION isPointNearTriangle(this, p)
			TYPE (Triangle),	INTENT (IN) :: this
			TYPE (Point),		INTENT (IN) :: p

!			IF (	(this%east >= getX(p))				.AND.							&
!					(getX(p) + width >= this%west)	.AND.							&
!					(this%north >= getY(p))				.AND.							&
!					(getY(p) + height >= this%south)	) THEN

			IF (getX(p) >= this%west	.AND.	getX(p) <= this%east	.AND.		&
			    getY(p) >= this%south	.AND.	getY(p) <= this%north)	THEN
				isPointNearTriangle = .TRUE.
			ELSE
				isPointNearTriangle = .FALSE.
			END IF
		END FUNCTION isPointNearTriangle

		!----------------------------------------------------------
		! Is a Point inside (or on) a Triangle?
		! If Point p is outside, then:
		! A(P,P1,P2) + A(P,P2,P3) + A(P,P3,P1) > A(P1,P2,P3)
		!
		! On input:
		!	this	-	Triangle
		!	p		-	Point to test
		!
		! On output:
		!	returns -	.TRUE.	if Point is inside (or on) Triangle
		!               .FALSE.	otherwise
		!----------------------------------------------------------
		LOGICAL FUNCTION isPointInTriangle(this, p)
			TYPE (Triangle),	INTENT (IN) :: this
			TYPE (Point),		INTENT (IN) :: p
			TYPE (Triangle)					:: t1, t2, t3
			REAL (8)								:: areaSum
			REAL (8)								:: areaRatio

			t1 = newTriangle(1, p, this%p1, this%p2)
			t2 = newTriangle(2, p, this%p2, this%p3)
			t3 = newTriangle(3, p, this%p3, this%p1)

			areaSum = getArea(t1) + getArea(t2) + getArea(t3)

!			IF (areaSum <= this%area) THEN
!				isPointInTriangle = .TRUE.
!			ELSE
!				isPointInTriangle = .FALSE.
!			END IF

			areaRatio = ABS(areaSum - this%area) / this%area


      		IF (areaRatio < small) THEN
				isPointInTriangle = .TRUE.
			ELSE
				isPointInTriangle = .FALSE.
			END IF

			CALL deleteTriangle(t3)
			CALL deleteTriangle(t2)
			CALL deleteTriangle(t1)
		END FUNCTION isPointInTriangle

		!--------------
		! Add a z-value
		!--------------
		SUBROUTINE addZ(this, z)
			TYPE (Triangle),	INTENT(IN OUT)	:: this
			REAL (8),			INTENT (IN)		:: z
			this%nz		= this%nz	+ 1					! Increment counter
			this%z		= this%z		+ z					! Increment sum(z)
			this%sdev	= this%sdev	+ z*z					! Increment sum(z**2)
		END SUBROUTINE addZ

		!--------------
		! Set a z-value
		!--------------
		SUBROUTINE setZee(this, z, nz, sdev)
			TYPE (Triangle),	INTENT(IN OUT)	:: this
			REAL (8), INTENT (IN)		:: z, sdev
			INTEGER, INTENT (IN) :: nz
			this%nz		= nz	! set counter
			this%z		= z	! set sum(z)
			this%sdev	= sdev	! set sdev
		END SUBROUTINE setZee

		!------------------------------------------------
		! Calculate average z-value over face of triangle
		!------------------------------------------------
		SUBROUTINE calcZ(this, envelope)
			TYPE (Triangle),	INTENT(IN OUT)				:: this
			LOGICAL,				INTENT (IN),	OPTIONAL	:: envelope

			IF (this%nz > 0) THEN
				!------------------
				! Compute average z
				!------------------
				this%z = this%z / this%nz

				!------------------------------------------------------
				! Compute standard deviation(z), a measure of roughness
				!------------------------------------------------------
				this%sdev = SQRT(MAX(this%sdev/this%nz - this%z*this%z, small))
				IF (this%sdev < small) this%sdev = zero

				!----------------------------------------------------
				! Add a specified amount of standard deviation to the
				! average value of z to obtain envelope topography.
				!----------------------------------------------------
!				IF (PRESENT(envelope) .AND. envelope) THEN
!					this%z = this%z + envWeight * this%sdev
!				END IF
			ELSE
				this%z		= zero
				this%sdev	= zero
			END IF
		END SUBROUTINE calcZ

		!----------------------
		! Print Triangle object
		!----------------------
		SUBROUTINE printTriangle(this)
			TYPE (Triangle), INTENT (IN) :: this
			WRITE (*, '("Triangle #", i7, ":")') this%number
			CALL printPoint(this%p1)
			CALL printPoint(this%p2)
			CALL printPoint(this%p3)
			WRITE (*, '("Area = ", f15.5)') this%area
			WRITE (*, '("Rectangular extent: ", /,									&
     &			"     +---------- north = ",  f12.3, " ---+", /,			&
     &			"     |                                   |", /,			&
     &			"     |                                   |", /,			&
     &			"west =",  f12.3, "     +            east =",f12.3,/,		&
     &			"     |                                   |", /,			&
     &			"     |                                   |", /,			&
     &			"     +---------- south = ",  f12.3, " ---+")')				&
     &			this%north, this%west, this%east, this%south
			WRITE (*, '("Width  = ", f12.3)') (this%east  - this%west )
			WRITE (*, '("Height = ", f12.3)') (this%north - this%south)
		END SUBROUTINE printTriangle

		!----------------------
		! Write Triangle object
		!----------------------
		SUBROUTINE writeTriangle(this, unit)
			TYPE (Triangle),	INTENT (IN)				:: this
			INTEGER,			INTENT (IN), OPTIONAL	:: unit
			IF (PRESENT(unit)) THEN
				WRITE (unit, '(i6,i5,3i7)')	this%number, 3,					&
						getID(this%p1),getID(this%p2),getID(this%p3)
			ELSE
				WRITE (*,    '(i6,i5,3i7)')	this%number, 3,					&
						getID(this%p1),getID(this%p2),getID(this%p3)
			END IF
		END SUBROUTINE writeTriangle

END MODULE TriangleClass
