#!/bin/sh -x
# --------------------------------------------------------------------------- #
# make_makefile.sh : Generates the makefile for WAVEWATCH based on switch     #
#                    settings                                                 #
#                    This script is called by w3_make and therefore does      #
#                    not print header information.                            #
#                                                                             #
# programs used : w3_new  Touches the correct files if compiler switches      #
#                         have been chsnged.                                  #
#                                                                             #
# error codes : all error output goes directly to screen in w3_make.          #
#                                                                             #
#                                                      Hendrik L. Tolman      #
#                                                      May 2002               #
# --------------------------------------------------------------------------- #
# 1. Preparations                                                             #
# --------------------------------------------------------------------------- #
# 1.a Internal variables

  ww3_env='.wwatch3.env'   # setup file

# 1.b ID header  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# 1.c Read data from the environment file  - - - - - - - - - - - - - - - - - -

  cd
  if [ -f $ww3_env ]
  then
    set `grep WWATCH3_DIR $ww3_env` ; shift
    main_dir="$*"
    set `grep WWATCH3_TMP $ww3_env` ; shift
    temp_dir="$*"
  else
     "*** Set-up file $ww3_env not found ***"
    exit 1
  fi

# 1.d Check / make directories   - - - - - - - - - - - - - - - - - - - - - - -

  if [ -d $main_dir ]
  then
    cd $main_dir
  else
    echo "*** Directory $main_dir not found ***"
    exit 2
  fi

  if [ ! -d $temp_dir ]
  then
    mkdir $temp_dir
  fi
  cd $temp_dir

# --------------------------------------------------------------------------- #
# 2. Part 1, subroutine dependencies                                          #
# --------------------------------------------------------------------------- #
# 2.a File ID

  echo ' '                                        > makefile
  echo '# WAVEWATCH III executables'             >> makefile
  echo '# -------------------------'             >> makefile

  rm -f filelist.tmp
  nr_thr=0

# 2.b Get info from switch file  - - - - - - - - - - - - - - - - - - - - - - -

  switch=$main_dir/bin/switch
  old_sw=$main_dir/bin/switch.old
  new_sw=$main_dir/bin/w3_new

  if [ ! -f $old_sw ]
  then
    old_sw=$switch
    $new_sw all
  fi

  for type in mach nco grib shared mpp thr0 thr1 c90 lrecl grid \
              prop source stab s_nl s_bot wind rwind curr \
              tpar tdyn dss0 pdif seed vamp
  do
    case $type in
      mach   ) TY='one'
               ID='hardware or compiler'
               OK='DUM F90' ;;
      nco    ) TY='upto1'
               ID='NCO modifications'
               TS='NCO'
               OK='NCO' ;;
      grib   ) TY='one'
               ID='GRIB package'
               OK='NOGRB NCEP1' ;;
      shared ) TY='one'
               ID='shared / distributed memory'
               OK='SHRD DIST' ;;
      mpp    ) TY='one'
               ID='message passing protocol'
               OK='SHRD MPI' ;;
      thr0   ) TY='upto1'
               ID='directive controlled threading (subs)'
               TS='OMP0'
               OK='OMP0' ;;
      thr1   ) TY='upto1'
               ID='directive controlled threading (loop)'
               TS='OMP1'
               OK='OMP1' ;;
      c90    ) TY='upto1'
               ID='Cray C90 compiler directives'
               TS='C90'
               OK='C90' ;;
      lrecl  ) TY='one'
               ID='word length in record length'
               OK='LRB4 LRB8' ;;
      grid   ) TY='one'
               ID='grid type (LL or XY)'
               OK='XYG LLG' ;;
      prop   ) TY='one'
               ID='propagation scheme'
               OK='PR0 PR1 PR2 PR3 PR4 PRX' ;;
      source ) TY='one'
               ID='source terms (input/diss)'
               OK='ST0 ST1 ST2 STX' ;;
      stab   ) TY='upto1'
               ID='stability correction'
               TS='STAB2'
               OK='STAB2' ;;
      s_nl   ) TY='one'
               ID='source terms (nonlinear int.)'
               OK='NL0 NL1 NL2 NLX' ;;
      s_bot  ) TY='one'
               ID='source terms (bottom friction)'
               OK='BT0 BT1 BTX' ;;
      wind   ) TY='one'
               ID='wind interpolation'
               OK='WND0 WND1 WND2' ;;
      rwind  ) TY='upto1'
               ID='wind vs. current definition'
               TS='RWND'
               OK='RWND' ;;
      curr   ) TY='one'
               ID='current interpolation'
               OK='CUR1 CUR2' ;;
      tpar   ) TY='upto1'
               ID='test parameter list output'
               TS='TPAR'
               OK='TPAR' ;;
      tdyn   ) TY='upto1'
               ID='dynamic diffusion time'
               TS='TDYN'
               OK='TDYN' ;;
      dss0   ) TY='upto1'
               ID='diffusion tensor'
               TS='DSS0'
               OK='DSS0' ;;
      pdif   ) TY='upto1'
               ID='propagation diffusion'
               TS='XW'
               OK='XW0 XW1' ;;
      seed   ) TY='upto1'
               ID='spectral seeding'
               TS='SEED'
               OK='SEED' ;;
      vamp   ) TY='upto1'
               ID='vampirtrace instrumentation'
               TS='VT'
               OK='VT' ;;
    esac

    n_found='0'
    s_found=
    for check in $OK
    do
      if [ "`grep $check $switch | wc -w | awk '{print $1}'`" -gt '1' ]
      then
        n_found=`expr $n_found + 1`
        s_found="$check $s_found"
      fi
    done

    if [ "$n_found" != '1' ] && [ "$TY" = 'one' ]
    then
      echo ' '
      echo "   *** No valid $ID switch found  ***"
      echo "       valid : $OK"
      echo "       found : $s_found"
      echo ' ' ; exit 3
    fi

    if [ "$n_found" -gt '1' ] && [ "$TY" = 'upto1' ]
    then
      echo ' '
      echo "   *** Too many $ID switches found (max 1) ***"
      echo "       valid : $OK"
      echo "       found : $s_found"
      echo ' ' ; exit 4
    fi

    if [ "$TY" = 'man' ]
    then
      echo "   w3_new $type has to be run manually"
    else
      if [ "$n_found" = '1' ]
      then
        sw="`echo $s_found | awk '{ print $1 }'`"
        if [ "`grep $sw $old_sw | wc -w | awk '{print $1}'`" -lt '1' ]
        then
          $new_sw $type
          echo "   new $ID"
        fi
      else
        if [ "`grep $TS $old_sw | wc -w | awk '{print $1}'`" -gt '1' ]
        then
          $new_sw $type
          echo "   new $ID"
        fi
      fi
    fi

    if [ "$n_found" = '1' ]
    then
      sw="`echo $s_found | awk '{ print $1 }'`"
    else
      sw=
    fi

    case $type in
      shared ) shared=$sw ;;
      mpp    ) mpp=$sw ;;
      thr0   ) nr_thr=`expr $nr_thr + $n_found` ;;
      thr1   ) nr_thr=`expr $nr_thr + $n_found` ;;
      prop   ) p_switch=$sw ;;
      source ) s_inds=$sw ;;
      stab   ) stab=$sw ;;
      s_nl   ) s_nl=$sw ;;
      s_bot  ) s_bt=$sw ;;
      vamp   ) vamp=$sw ;;
        *    ) ;;
    esac
  done

  if [ "$nr_thr" != '0' ] && [ "$shared" != 'SHRD' ]
  then
      echo ' '
      echo "   *** !/OMPn has to be used in combination with !/SHRD"
      echo ' ' ; exit 5
  fi

  case $p_switch in
   PR0) pr=$NULL               ; prx=$NULL               ;;
   PR1) pr='w3pro1md'          ; prx=$NULL               ;;
   PR2) pr='w3uqckmd w3pro2md' ; prx='w3uqckmd w3pro2md' ;;
   PR3) pr='w3uqckmd w3pro3md' ; prx='w3uqckmd w3pro3md' ;;
   PR4) pr='w3uqckmd w3pro4md' ; prx='w3uqckmd w3pro4md' ;;
   PRX) pr='w3proxmd'          ; prx=$NULL               ;;
  esac

  case $s_inds in
   ST0) st=$NULL ;;
   ST1) st='w3src1md' ;;
   ST2) st='w3src2md' ;;
   STX) st='w3srcxmd' ;;
  esac

  if [ "$stab" = 'STAB2' ] && [ "$s_inds" != 'ST2' ]
  then
      echo ' '
      echo "   *** !/STAB2 has to be used in combination with !/ST2"
      echo ' ' ; exit 6
  fi

  case $s_nl in
   NL0) nl=$NULL ;;
   NL1) nl='w3snl1md' ;;
   NL2) nl='w3snl2md mod_xnl4v4 serv_xnl4v4 mod_constants mod_fileio' ;;
   NLX) nl='w3snlxmd' ;;
  esac

  case $s_bt in
   BT0) bt=$NULL ;;
   BT1) bt='w3sbt1md' ;;
   BTX) bt='w3sbtxmd' ;;
  esac

  if [ "$vamp" = 'VT' ] && [ "$mpp" != 'MPI' ]
  then
      echo ' '
      echo "   *** !/VT has to be used in combination with !/MPI"
      echo ' ' ; exit 7
  fi

  if [ "$nr_thr" != '0' ] && [ "$s_nl" = 'NL2' ]
  then
      echo ' '
      echo "   *** The present version of the WRT interactions"
      echo "       cannot be run under OpenMP (OMP0, OMP1). Use"
      echo "       SHRD or MPI options instead.                    ***"
      echo ' ' ; exit 8
  fi

# 2.c Make makefile and file list  - - - - - - - - - - - - - - - - - - - - - -

  progs='ww3_grid ww3_strt ww3_prep ww3_shel ww3_outf ww3_outp ww3_trck'
  progs="$progs ww3_grib gx_outf gx_outp"

  for prog in $progs
  do
    case $prog in
     ww3_grid) IDstring='Grid preprocessor'
               core=
               prop="$prx"
             source="w3srcemd $st $nl $bt"
                 IO='w3iobcmd w3iogrmd'
                upd=
                aux="constants w3parmmd w3testmd w3servmd w3dispmd w3arrymd" ;;
     ww3_strt) IDstring='Initial conditions program'
               core=
               prop="$prx"
             source="w3srcemd $st $nl $bt"
                 IO='w3iorsmd w3iobcmd w3iogrmd'
                upd=
                aux="constants w3parmmd w3testmd w3servmd w3dispmd w3arrymd" ;;
     ww3_prep) IDstring='Field preprocessor'
               core='w3fldsmd'
               prop="$prx"
             source="w3srcemd $st $nl $bt"
                 IO='w3iogrmd w3iobcmd'
                upd=
                aux="constants w3parmmd w3testmd w3servmd w3dispmd w3timemd w3arrymd" ;;
     ww3_shel) IDstring='Generic shell'
               core='w3fldsmd w3wavemd w3wdasmd'
               prop="$pr"
             source="w3srcemd $st $nl $bt"
                 IO='w3iogrmd w3iogomd w3iopomd w3iotrmd w3iorsmd w3iobcmd'
                upd='w3updtmd'
                aux="constants w3parmmd w3testmd w3servmd w3dispmd w3timemd w3arrymd " ;;
     ww3_outf) IDstring='Gridded output'
               core=
               prop="$prx"
             source="w3srcemd $st $nl $bt"
                 IO='w3iogrmd w3iogomd w3iobcmd'
                upd=
                aux="constants w3parmmd w3testmd w3servmd w3dispmd w3arrymd w3timemd" ;;
     ww3_outp) IDstring='Point output'
               core=
               prop="$prx"
             source="w3srcemd $st $nl $bt"
                 IO='w3iogrmd w3iopomd w3iobcmd'
                upd=
                aux="constants w3parmmd w3testmd w3servmd w3dispmd w3timemd w3arrymd" ;;
     ww3_trck) IDstring='Track output post'
               core=
               prop=
             source=
                 IO=
                upd=
                aux="w3servmd w3timemd" ;;
     ww3_grib) IDstring='Gridded output (GRIB)'
               core=
               prop="$prx"
             source="w3srcemd $st $nl $bt"
                 IO='w3iogrmd w3iogomd w3iobcmd'
                upd=
                aux="constants w3parmmd w3testmd w3servmd w3dispmd w3timemd" ;;
      gx_outf) IDstring='GrADS input file generation (gridded fields)'
               core=
               prop="$prx"
             source="w3srcemd $st $nl $bt"
                 IO='w3iogrmd w3iogomd w3iobcmd'
                upd=
                aux="constants w3parmmd w3testmd w3servmd w3dispmd w3timemd" ;;
      gx_outp) IDstring='GrADS input file generation for point output'
               core=
               prop="$prx"
             source="w3srcemd $st $nl $bt"
                 IO='w3iogrmd w3iopomd w3iobcmd'
                upd=
                aux="constants w3parmmd w3testmd w3servmd w3dispmd w3timemd" ;;
    esac

    d_string='$(aPe)/'"$prog"' : $(aPo)/'
    files="$core $prop $source $IO $upd $aux $prog"
    filesl="$prog $core $prop $source $IO $upd $aux"

    echo "# $IDstring"                           >> makefile
    echo ' '                                     >> makefile
    for file in $files
    do
      echo "$d_string$file.o"                    >> makefile
      echo "$file"                               >> filelist.tmp
    done
    echo '	@$(aPb)/link '"$filesl"          >> makefile
    echo ' '                                     >> makefile
  done

  sort -u filelist.tmp                            > filelist
  rm -f filelist.tmp

# --------------------------------------------------------------------------- #
# 3. Part 2, includes in subroutines                                          #
# --------------------------------------------------------------------------- #
# 3.a File ID

  echo '   Checking all subroutines for modules (this may take a while) ...'

  echo ' '                                       >> makefile
  echo '# WAVEWATCH III subroutines'             >> makefile
  echo '# -------------------------'             >> makefile
  echo ' '                                       >> makefile

# 3.b Loop over files

  for file in `cat filelist`
  do
    if [ -f $main_dir/ftn/$file.ftn ]
    then
      fext=ftn
    else
      if [ -f $main_dir/ftn/$file.f90 ]
      then
        fext=f90
      else
        fext=f
      fi
    fi

    string1='$(aPo)/'$file'.o : '$file.$fext' '
    string2='	@$(aPb)/ad3'" $file"
    string3="$NULL"

    ad3 $file 0 1 > ad3.out 2>&1
    if [ -n "`grep error ad3.out`" ]
    then
      cat ad3.out
      exit 8
    fi
    rm -f ad3.out

    if [ -f $file.f90 ]
    then
      fext=f90
    else
      fext=f
    fi

    grep USE $file.$fext  > check_file
    grep use $file.$fext >> check_file
    rm -f $file.$fext

    for mod in CONSTANTS W3PARMMD W3TESTMD W3WAVEMD W3WDASMD W3UPDTMD \
               W3PRO1MD W3PRO2MD W3PRO3MD W3PRO4MD W3PROXMD W3UQCKMD \
               W3SRCEMD W3SRC1MD W3SRC2MD W3SRCXMD \
               W3SNL1MD W3SNL2MD W3SNLXMD \
               m_xnldata serv_xnl4v4 m_fileio m_constants \
               W3SBT1MD W3SBTXMD \
               W3IOGRMD W3IOGOMD W3IOPOMD W3IOTRMD W3IORSMD W3IOBCMD \
               W3DISPMD W3TIMEMD W3SERVMD W3ARRYMD W3FLDSMD
    do
      case $mod in
         'CONSTANTS'    ) modtest=constants.o ;;
         'W3PARMMD'     ) modtest=w3parmmd.o ;;
         'W3TESTMD'     ) modtest=w3testmd.o ;;
         'W3WAVEMD'     ) modtest=w3wavemd.o ;;
         'W3WDASMD'     ) modtest=w3wdasmd.o ;;
         'W3UPDTMD'     ) modtest=w3updtmd.o ;;
         'W3PRO1MD'     ) modtest=w3pro1md.o ;;
         'W3PRO2MD'     ) modtest=w3pro2md.o ;;
         'W3PRO3MD'     ) modtest=w3pro3md.o ;;
         'W3PRO4MD'     ) modtest=w3pro4md.o ;;
         'W3UQCKMD'     ) modtest=w3uqckmd.o ;;
         'W3PROXMD'     ) modtest=w3proxmd.o ;;
         'W3SRCEMD'     ) modtest=w3srcemd.o ;;
         'W3SRC1MD'     ) modtest=w3src1md.o ;;
         'W3SRC2MD'     ) modtest=w3src2md.o ;;
         'W3SRCXMD'     ) modtest=w3srcxmd.o ;;
         'W3SNL1MD'     ) modtest=w3snl1md.o ;;
         'W3SNL2MD'     ) modtest=w3snl2md.o ;;
         'm_xnldata'    ) modtest=mod_xnl4v4.o ;;
         'serv_xnl4v4'  ) modtest=serv_xnl4v4.o ;;
         'm_fileio'     ) modtest=mod_fileio.o ;;
         'm_constants'  ) modtest=mod_constants.o ;;
         'W3SNLXMD'     ) modtest=w3snlxmd.o ;;
         'W3SBT1MD'     ) modtest=w3sbt1md.o ;;
         'W3SBTXMD'     ) modtest=w3sbtxmd.o ;;
         'W3IOGRMD'     ) modtest=w3iogrmd.o ;;
         'W3IOGOMD'     ) modtest=w3iogomd.o ;;
         'W3IOPOMD'     ) modtest=w3iopomd.o ;;
         'W3IOTRMD'     ) modtest=w3iotrmd.o ;;
         'W3IORSMD'     ) modtest=w3iorsmd.o ;;
         'W3IOBCMD'     ) modtest=w3iobcmd.o ;;
         'W3DISPMD'     ) modtest=w3dispmd.o ;;
         'W3TIMEMD'     ) modtest=w3timemd.o ;;
         'W3SERVMD'     ) modtest=w3servmd.o ;;
         'W3ARRYMD'     ) modtest=w3arrymd.o ;;
         'W3FLDSMD'     ) modtest=w3fldsmd.o ;;
      esac
      nr=`grep $mod check_file | wc -c | awk '{ print $1 }'`
      if [ "$nr" -gt '8' ]
      then
        if [ "$modtest" != "$file.o" ]
        then
          string3="$string3 "'$(aPo)/'"$modtest"
        fi
      fi
    done
    rm -f check_file

    echo "$string1$string3"                      >> makefile
    echo "$string2"                              >> makefile
    echo ' '                                     >> makefile

  done

  echo '# end of WAVEWATCH III subroutines'      >> makefile
  rm -f filelist

# --------------------------------------------------------------------------- #
# 4. Move makefile to proper place                                            #
# --------------------------------------------------------------------------- #

  mv makefile $main_dir/ftn/makefile

# end of script ------------------------------------------------------------- #
